import uvicore
from uvicore.package import Provider
from uvicore.support.dumper import dump, dd
from uvicore.redis.package import bootstrap
from uvicore.foundation.events import app as AppEvents
from uvicore.redis.package.registers import Redis as RedisMixin


@uvicore.provider()
class Redis(Provider, RedisMixin):

    def register(self) -> None:
        """Register package into the uvicore framework.
        All packages are registered before the framework boots.  This is where
        you define your packages configs, IoC bindings and early event listeners.
        Configs are deep merged only after all packages are registered.  No real
        work should be performed here as it is very early in the bootstraping
        process and we have no clear view of the full configuration system."""

        # Register event listeners
        AppEvents.Booted.listen(bootstrap.Redis)

        # Event Handlers
        # String based events instead of class based because HTTP may not even
        # be installed, so importing the HTTP event would cause an issue.

        # Disconnect from all redis pools after the system has shutdown
        @uvicore.events.handle(['uvicore.console.events.command.Shutdown', 'uvicore.http.events.server.Shutdown'])
        async def uvicore_shutdown(event):
            # Trying to solve this error when I do a huge wrk test
            # Task was destroyed but it is pending!
            # task: <Task pending name='Task-13' coro=<RedisConnection._read_data() running at /home/mreschke/.cache/pypoetry/virtualenvs/mreschke-speedtest-epfwGmSK-py3.9/lib/python3.9/site-packages/aioredis/connection.py:186> wait_for=<Future pending cb=[<TaskWakeupMethWrapper object at 0x7f6043f7ac10>()]> cb=[RedisConnection.__init__.<locals>.<lambda>() at /home/mreschke/.cache/pypoetry/virtualenvs/mreschke-speedtest-epfwGmSK-py3.9/lib/python3.9/site-packages/aioredis/connection.py:168]>
            # See https://github.com/aio-libs/aioredis-py/issues/154
            # I am doing the proper engine.close() and await engine.wait_closed() and it does run before I get the error
            # but the error persists.  Still not fully solved, but closing does help in some situations
            from uvicore.redis.redis import Redis as redis
            for engine in redis.engines.values():
                engine.close()
                await engine.wait_closed()

    def boot(self) -> None:
        """Bootstrap package into the uvicore framework.
        Boot takes place after ALL packages are registered.  This means all package
        configs are deep merged to provide a complete and accurate view of all
        configuration. This is where you register, connections, models,
        views, assets, routes, commands...  If you need to perform work after ALL
        packages have booted, use the event system and listen to the booted event:
        self.events.listen('uvicore.foundation.events.app.Booted', self.booted)"""

        pass
        # Define service provider registration control
        #self.registers(self.package.config.registers)

        # Define Redis Connections
        # self.redis_connections(
        #     connections=self.package.config.redis.connections,
        #     default=self.package.config.redis.default
        # )



import uvicore
from uvicore.typing import Dict
from uvicore.package import Provider
from uvicore.support.dumper import dump, dd
from uvicore.console.package.registers import Cli
from uvicore.support.module import load
from uvicore.console import group as cli_group
from uvicore.console.package import bootstrap
from uvicore.foundation.events import app as AppEvents


@uvicore.provider()
class Console(Provider, Cli):

    def register(self) -> None:
        """Register package into the uvicore framework.
        All packages are registered before the framework boots.  This is where
        you define your packages configs, IoC bindings and early event listeners.
        Configs are deep merged only after all packages are registered.  No real
        work should be performed here as it is very early in the bootstraping
        process and we have no clear view of the full configuration system."""

        # Register IoC bindings
        #from uvicore.console.console import cli
        #dump('service------------')
        # self.bind('uvicore.console.console.cli', 'uvicore.console.console.cli',
        #     aliases=['Console', 'console', 'cli', 'cli2']
        # )

        # self.bind('Console', 'uvicore.console.console.cli',
        #     aliases=['uvicore.console.console.cli', 'console', 'cli', 'cli2']
        # )

        # Register event listeners
        # Priority is 90 because we want the console to be bootstrapped after most
        # other uvicore services like Database.  If database is not initialized, and you import
        # a DB model/table from a command, it will error because the connection strings have not yet
        # been initialized.
        AppEvents.Booted.listen(bootstrap.Console, priority=90)
        #uvicore.events.listen('uvicore.foundation.events.app.Booted', bootstrap.Console, priority=90)
        #uvicore.events.listen('uvicore.foundation.events.app.*', bootstrap.Console, priority=90)

    def boot(self) -> None:
        """Bootstrap package into the uvicore framework.
        Boot takes place after ALL packages are registered.  This means all package
        configs are deep merged to provide a complete and accurate view of all
        configuration. This is where you register, connections, models,
        views, assets, routes, commands...  If you need to perform work after ALL
        packages have booted, use the event system and listen to the booted event:
        self.events.listen('uvicore.foundation.events.app.Booted', self.booted)"""

        # Define service provider registration control
        # No - Never allow this packages registrations to be disabled from other configs

        # Define commands
        self.register_cli_commands(
            group='gen',
            help='Generate New Schematics (commands, models, views...)',
            commands={
                'command': 'uvicore.console.commands.generators.command'
            }
        )



import re
import uvicore
import inspect
import asyncio
from uvicore.typing import Dict, List, Any, Union, Callable, Tuple
from uvicore.support.dumper import dump, dd
from types import SimpleNamespace as obj
from collections import namedtuple
from uvicore.contracts import Dispatcher as DispatcherInterface
from uvicore.support import module
from uvicore.support.concurrency import run_in_threadpool

#from uvicore.contracts import Event as EventInterface
# from prettyprinter import pretty_call, register_pretty
# class EventInfo(EventInterface):
#     """Event Definition"""
#     # These class level properties for for type annotations only.
#     # They do not restrict of define valid properties like a dataclass would.
#     # This is still a fully dynamic SuperDict!
#     name: str
#     description: str
#     dynamic: bool
#     is_async: bool
#     options: Dict


# @register_pretty(EventInfo)
# def pretty_entity(value, ctx):
#     """Custom pretty printer for my SuperDict"""
#     # This printer removes the class name uvicore.types.Dict and makes it print
#     # with a regular {.  This really cleans up the output!

#     #return pretty_call(ctx, 'Dict', **value)
#     #return pretty_call(ctx, 'Dict', value.to_dict())  # Does show nested dicts as SuperDicts
#     return pretty_call(ctx, 'EventInfo', dict(**value))

@uvicore.service('uvicore.events.dispatcher.Dispatcher',
    aliases=['Dispatcher', 'dispatcher', 'Event', 'event'],
    singleton=True,
)
class Dispatcher(DispatcherInterface):
    """Event Dispatcher private class.

    Do not import from this location.
    Use the uvicore.events singleton global instead."""

    # @property
    # def events(self) -> Dict[str, EventInfo]:
    #     """Dictionary of all registered events in uvicore and all packages"""
    #     return self._events

    @property
    def listeners(self) -> Dict[str, List]:
        """Dictionary of all listeners for each event"""
        return self._listeners

    @property
    def wildcards(self) -> List:
        """List of all wildcard listeners"""
        return self._wildcards

    def __init__(self) -> None:
        self._events: Dict[str, EventInfo] = Dict()
        self._listeners: Dict[str, List] = Dict()
        self._wildcards: List = []

    @property
    def registered_events(self) -> List:
        """Get all registered events from IOC bindings and manual registrations"""
        # FIXME, need to merge in manual registrations, which I don't have yet
        event_bindings = uvicore.ioc.binding(type='event')
        events = []
        for binding in event_bindings.values():
            events.append({
                'name': binding.path,
                'description': binding.object.__doc__,
                'is_async': binding.object.is_async,
            })
        return events


    # def event(self, event: Union[str, Callable]) -> EventInfo:
    #     """Get one known (pre-registered) EventInfo by str name or class"""
    #     if type(event) == str:
    #         # Get event by a string name
    #         name = event
    #     else:
    #         # Get event by a class, use the class name as the event string name
    #         #name = str(event.__class__).split("'")[1]
    #         name = event.name
    #     if name in self.events:
    #         return self.events.get(name)
    #     else:
    #         # This event is NOT registered, but we still want it to work
    #         # because of all the dynamic events like ORM makes
    #         # So create a fake event meta
    #         return Dict({
    #             'name': name
    #         })

    @property
    def expanded_sorted_listeners(self) -> List:
        """Get all listeners with expanded wildcards, sorted by priority ASC"""

        # WELL, this doesn't really work because it only merged wildcards into
        # events that are already explicitly listened too
        # like uvicore.foundation.events.app.* only shows up under Booted
        # not Registered because no other event has explicitly listened to
        # Registered.  I could look up self.registered_events as well
        # but this is really only for the CLI event listeners, so maybe Ill skip

        listeners = {}
        for event, listener in self.listeners.items():
            if '*' not in event:
                listeners[event] = listener

                for wildcard in self.wildcards:
                    if re.search(wildcard, event):
                        if event not in listeners:
                            listeners[event] = []
                        listeners[event].extend(self.listeners[wildcard])
                        break

        return listeners

    def event_listeners(self, event: str) -> List:
        """Get all listeners for an event including wildcard, sorted by priority ASC"""

        # Get all listeners for this particular event
        listeners = [x for x in self.listeners.get(event) or []]

        # Add in wildcard events
        for wildcard in self.wildcards:
            regex = wildcard
            if re.search(regex, event):
                listeners.extend(self.listeners[wildcard])

        # Sort listeners by priority
        listeners = sorted(listeners, key = lambda i: i['priority'])

        # Get just the listener handler strings/methods as List
        handlers = [x['listener'] for x in listeners]

        # Return these event handlers
        return handlers

    def listen(self, events: Union[str, List], listener: Union[str, Callable] = None, *, priority: int = 50) -> None:
        """Decorator or method to append a listener (string or Callable) callback to one or more events."""
        def handle(events, listener):
            if type(events) != list: events = [events]

            # Expand all wildcards
            for event in events:
                # Get the string name of this event
                if type(event) != str: event = event.name

                # If event not registered yet, ensure empty []
                if event not in self.listeners:
                    self._listeners[event] = []

                # Append new listener to event
                self._listeners[event].append({'listener': listener, 'priority': priority})

                # If event contains a *, add it to our wildcard list for use later
                if '*' in event:
                    self._wildcards.append(event)

        # Method access
        if listener: return handle(events, listener)

        # Decorator access
        def decorator(func):
            handle(events, func)
            return func
        return decorator

    def handle(self, events: Union[str, List], listener: Union[str, Callable] = None, *, priority: int = 50) -> None:
        """Decorator or method to append a listener (string or Callable) callback to one or more events.  Alias to listen()."""
        return self.listen(events, listener)

    def subscribe(self, listener: Union[str, Callable]) -> None:
        """Add a subscription class which handles both registration and listener callbacks"""
        try:
            if type(listener) == str:
                module.load(listener).object().subscribe(uvicore.events)
            else:
                listener.subscribe(uvicore.events)
        except ModuleNotFoundError:
            pass

    def dispatch(self, event: Union[str, Callable], payload: Dict = {}) -> None:
        """Fire off an event and run all listener callbacks"""

        # Get dispatcher method for this event
        dispatch_method, params = self._get_dispatcher(event, payload, is_async=False)

        # Dispatch this event
        dispatch_method(*params)

    async def dispatch_async(self, event: Union[str, Callable], payload: Dict = {}) -> None:
        """Async fire off an event and run all async listener callbacks"""

        # Get dispatcher method for this event
        dispatch_method, params = self._get_dispatcher(event, payload, is_async=True)

        # Dispatch this event
        await dispatch_method(*params)

    async def codispatch(self, event: Union[str, Callable], payload: Dict = {}) -> None:
        """Alias for dispatch_async()."""
        return await self.dispatch_async(event, payload)

    def _dispatch(self, event: Union[str, Callable], payload: Dict = {}) -> None:
        """Dispatch an event by fireing off all listeners/handlers"""
        (event, handlers) = self._get_handlers(event, payload)
        for handler in handlers:
            handler(event)

    async def _dispatch_async(self, event: Union[str, Callable], payload: Dict = {}) -> None:
        """Dispatch an event by fireing off all listeners/handlers"""
        (payload, handlers) = self._get_handlers(event, payload)
        for handler in handlers:
            if asyncio.iscoroutinefunction(handler) or asyncio.iscoroutinefunction(handler.__call__):
                await handler(event)
            else:
                # Listener/handler is NOT async but was called from await, lets throw in thread pool
                await run_in_threadpool(handler, event)

    def _get_dispatcher(self, event: Union[str, Callable], payload: Dict = {}, is_async: bool = False) -> Tuple:
        """Get dispatcher method for this event"""
        # This function determines the proper dispatching method for this event.
        # If event is a string, we use self._dispatch
        # If event is a class we call that classes .dispatch method
        method = None
        params = [event, payload]
        if type(event) == str:
            if '-' in event or '{' in event:
                # String event (we know because classes can't have dashes or {.
                # This is how we name dynamic string based events like per model or table...
                method = self._dispatch_async if is_async else self.dispatch
            else:
                # See if string event has a matching class.  If so, import and dispatch it
                try:
                    event = module.load(event).object(**payload)
                    params = []
                    method = event.dispatch_async if is_async else event.dispatch
                except:
                    # No class found for this string.  This is OK because events can
                    # be strings without matching classes.  Dispatch it anyway
                    method = self._dispatch_async if is_async else self._dispatch
        else:
            # Event is an event class INSTANCE.  Call the actual classes dispatch method
            # in case the user overrode it, we still execute it
            params = []
            method = event.dispatch_async if is_async else event.dispatch

        # Return tuple of dispatcher method and params
        return (method, params)

    def _get_handlers(self, event: Union[str, Callable], payload: Dict = {}) -> Tuple:
        """Get all listener/handlers and fix up payload"""
        # Get event by string name or class inspection
        #event_meta = self.event(event)

        if type(event) == str:
            # String based event, merge payload with default event
            event = Dict(payload).merge({
                'name': event,
                'description': 'String based dynamic event.'
            })

        # Payload default is a SuperDict
        #payload = Dict(payload)

        # If event is a class, payload is the class instance
        #if type(event) != str: payload = event

        # Get listener methods (dynamic import if string)
        listeners = self.event_listeners(event.name)
        handlers = []
        for handler in listeners:
            if type(handler) == str:
                # handler is a string to a listener class with handle() method
                try:
                    #cls = module.load(listener).object(uvicore.app)
                    cls = module.load(handler).object()
                    handlers.append(cls)
                    #handlers.append(cls.handle)
                except ModuleNotFoundError:
                    # Bad handler, handler will never fire
                    continue
            else:
                # handler is a Class
                if inspect.isclass(handler): handler = handler()

                # Listener is a Callable (if was a class, NOW its callable)
                handlers.append(handler)

        # Return tuple of event and handlers
        #return (event_meta, payload, handlers)
        return (event, handlers)


# IoC Class Instance
# **Not meant to be imported from here**.  Use the uvicore.events singleton global instead.
# Only here because uvicore bootstrap needs to import it without a service provider.
# By using the default bind and make feature of the IoC we can swap the implimentation
# at a high bootstrap level using our app configs 'bindings' dictionary.
# The only two classes that do this are Application and the event Dispatcher.
#Dispatcher: _Dispatcher = uvicore.ioc.make('Dispatcher', _Dispatcher, singleton=True, aliases=['dispatcher', 'Event', 'event', 'Events', 'events'])



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Union, Mapping, Optional

try:
    from sqlalchemy.engine import Engine
    from sqlalchemy import MetaData, Table
    from databases import Database as EncodeDatabase
    from sqlalchemy.sql import ClauseElement
    from sqlalchemy.engine.result import RowProxy
except:
    Engine = None
    Table = None
    MetaData = None
    EncodeDatabase = None
    ClauseElement = None
    RowProxy = None

from uvicore.contracts import DbQueryBuilder
from .connection import Connection
from .package import Package


class Database(ABC):
    pass

    @property
    @abstractmethod
    def default(self) -> str:
        """The default connection str for the main running app"""
        pass

    @property
    @abstractmethod
    def connections(self) -> Dict[str, Connection]:
        """All connections from all packages, keyed by connection str name"""
        pass

    @property
    @abstractmethod
    def engines(self) -> Dict[str, Engine]:
        """All engines for all unique (by metakey) connections, keyed by metakey"""
        pass

    @property
    @abstractmethod
    def databases(self) -> Dict[str, EncodeDatabase]:
        """All Encode Databases for all unique (by metakey) connections, keyed by metakey"""
        pass

    @property
    @abstractmethod
    def metadatas(self) -> Dict[str, MetaData]:
        """All SQLAlchemy Metadata for all unique (by metakey) connections, keyed by metakey"""
        pass

    @abstractmethod
    def init(self, default: str, connections: List[Connection]) -> None:
        """Initialize the database system with a default connection str and List of all Connections from all packages"""
        pass

    @abstractmethod
    def packages(self, connection: str = None, metakey: str = None) -> List[Package]:
        """Get all packages with the metakey (direct or derived from connection str)."""
        pass

    @abstractmethod
    def metakey(self, connection: str = None, metakey: str = None) -> str:
        """Get one metekay by connection str or metakey"""
        pass

    @abstractmethod
    def connection(self, connection: str = None) -> Connection:
        """Get one connection by connection name"""
        pass

    @abstractmethod
    def metadata(self, connection: str = None, metakey: str = None) -> MetaData:
        """Get one SQLAlchemy Metadata by connection str or metakey"""
        pass

    @abstractmethod
    def tables(self, connection: str = None, metakey: str = None) -> List[Table]:
        """Get all tables for a given connection or metakey"""
        pass

    @abstractmethod
    def table(self, table: str, connection: str = None) -> Table:
        """Get one SQLAlchemy Table by name (without prefix) and connection str or connection.tablename dot notation"""
        pass

    @abstractmethod
    def tablename(self, table: str, connection: str = None) -> str:
        """Get a SQLAlchemy tablename with prefix by name (without prefix) and connection str or connection.tablename dot notation"""
        pass

    @abstractmethod
    def engine(self, connection: str = None, metakey: str = None) -> Engine:
        """Get one SQLAlchemy Engine by connection str or metakey"""
        pass

    @abstractmethod
    async def database(self, connection: str = None, metakey: str = None) -> EncodeDatabase:
        """Get one Encode Database by connection str or metakey"""
        pass

    @abstractmethod
    async def connect(self, connection: str = None, metakey: str = None, *, all_dbs: bool = False) -> None:
        """Connect to a database by connection str or metakey.  Or connect ALL databases."""

    @abstractmethod
    async def disconnect(self, connection: str = None, metakey: str = None, all_dbs: bool = False) -> None:
        """Disconnect from a database by connection str or metakey.  Or disconnect ALL databases."""
        pass

    @abstractmethod
    async def fetchall(self, query: Union[ClauseElement, str], values: Dict = None, connection: str = None, metakey: str = None) -> List[RowProxy]:
        """Fetch List of records from a SQLAlchemy Core Query based on connection str or metakey"""
        pass

    @abstractmethod
    async def fetchone(self, query: Union[ClauseElement, str], values: Dict = None, connection: str = None, metakey: str = None) -> Optional[RowProxy]:
        """Fetch one record from a SQLAlchemy Core Query based on connection str or metakey"""
        pass

    @abstractmethod
    async def execute(self, query: Union[ClauseElement, str], values: Union[List, Dict] = None, connection: str = None, metakey: str = None) -> Any:
        """Execute a SQLAlchemy Core Query based on connection str or metakey"""
        pass

    @abstractmethod
    def query(self, connection: str = None) -> DbQueryBuilder[DbQueryBuilder, None]:
        """Database query builder passthrough"""

    # @property
    # @abstractmethod
    # def events(self) -> Dict: pass

    # @property
    # @abstractmethod
    # def listeners(self) -> Dict[str, List]: pass

    # @property
    # @abstractmethod
    # def wildcards(self) -> List: pass

    # @abstractmethod
    # def register(self, events: Dict):
    #     pass

    # @abstractmethod
    # def listen(self, events: Union[str, List], listener: Any) -> None:
    #     pass

    # @abstractmethod
    # def dispatch(self, event: Any, payload = {}) -> None:
    #     pass

    # @abstractmethod
    # def get_listeners(self, event: str) -> List:
    #     pass



from abc import ABC, abstractmethod
from dataclasses import dataclass
from uvicore.typing import Any, Callable, Dict, List, Optional, TypeVar, Union

T = TypeVar('T')


@dataclass
class Binding(ABC):
    path: Optional[str]
    object: Optional[Any]
    instance: Optional[Any]
    type: Optional[str]
    factory: Optional[Any]
    kwargs: Optional[Dict]
    singleton: bool
    aliases: List


class Ioc(ABC):
    @property
    @staticmethod
    def bindings(self) -> Dict[str, Binding]:
        pass

    @property
    @staticmethod
    def aliases(self) -> Dict[str, str]:
        pass

    # @abstractmethod
    # def config(self, config: Dict) -> None:
    #     """Set the main running app config dictionary for IoC binding override configs"""
    #     pass

    @abstractmethod
    def binding(self, name: str = None, *, type: str = None, include_overrides: bool = True) -> Union[Binding, Dict]:
        """Get an IoC binding object by name or alias"""
        pass

    @abstractmethod
    def make(self, name: str, default: Callable[[], T] = None, **kwargs) -> T:
        """Make a module/class/method by name or alias from IoC bindings"""
        pass

    @abstractmethod
    def bind_from_decorator(self, cls, name: str = None, *, object_type: str = None, factory: Any = None, kwargs: Dict = None, singleton: bool = False, aliases: List = []) -> None:
        """Bind from a decorator"""
        pass

    @abstractmethod
    def bind_override(self, name: str, object: str):
        """Add a binding override to an array to check later"""
        pass

    @abstractmethod
    def bind(self, name: str = None, object: Any = None, *, object_type: str = 'service', override: bool = True, factory: Any = None, kwargs: Dict = None, singleton: bool = False, aliases: List = []) -> None:
        """Add an IoC binding"""
        pass

    @abstractmethod
    def bind_map(self, mapping: Dict[str, Dict]) -> None:
        """Add IoC bindings from dictionary"""
        pass

    @abstractmethod
    def alias(self, src: str, dest: str) -> None:
        """Add alias to existing binding"""
        pass



import uvicore
from uvicore import log
from uvicore.support.dumper import dd, dump
from uvicore.console import command, argument

# Commands
#list = typer.Typer()
#show = typer.Typer()


@command()
def list():
    """List all events"""
    log.header("Events defined from all packages")
    log.line()

    event_bindings = uvicore.ioc.binding(type='event')
    events = []
    for binding in event_bindings.values():
        events.append({
            'name': binding.path,
            'description': binding.object.__doc__,
            'is_async': binding.object.is_async,
        })
    dump(events)


@command()
@argument('event')
def get(event: str):
    """Show detailed info for one event"""
    log.header("Event details for " + event)
    log.line()

    event_bindings = uvicore.ioc.binding(type='event')
    events = []
    for binding in event_bindings.values():
        if event == binding.path:
            events.append({
                'name': binding.path,
                'description': binding.object.__doc__,
                'is_async': binding.object.is_async,
            })
    if events:
        dump(events)
    else:
        print("Event {} not found".format(event))


@command()
def listeners():
    """Show all event listeners/handlers"""
    log.header("Event listeners/handlers")
    log.line()

    #dump(uvicore.events.expanded_sorted_listeners)
    dump(uvicore.events.listeners)



from abc import ABC, abstractmethod, abstractproperty
from uvicore.typing import Any, Dict, List, Tuple, OrderedDict, Union
from .config import Config
from .server import Server
from .package import Package
from .template import Template

try:
    from fastapi import FastAPI
    from starlette.applications import Starlette
except ImportError:  # pragma: nocover
    FastAPI = None  # type: ignore
    Starlette = None  # type: ignore

# Regular attributes (version: str = '0.0.0) vs
# @property @abstractmethod

# Defining attributes here does MAKE the attributes in the
# inheriting class implicitly.  But we don't really want that
# because the user now accept the abstrac classes defaults
# and is not forced to impliment them at all.  We want to FORCE
# the implimentation to define their own attribute values.

# So @property @abstractmethod is the best way to enforce
# requirements and conformity.

class Application(ABC):
    """Main uvicore application"""

    # Instance Variables
    @abstractproperty
    def version(self) -> str:
        """Uvicore framework version number"""
        pass

    @abstractproperty
    def debug(self) -> bool:
        """Debug mode for entire application"""
        pass

    @abstractproperty
    def perfs(self) -> List:
        """List of all perf dumps for performance tuning"""
        pass

    @abstractproperty
    def http(self) -> Union[Starlette, FastAPI]:
        """HTTP Server Instance"""
        pass

    # @property
    # @abstractmethod
    # def config(self) -> Config:
    #     """Configuration system"""
    #     pass

    @abstractproperty
    def providers(self) -> OrderedDict[str, Dict]:
        """OrderedDict of providers from all packages in proper dependency order"""
        pass

    @abstractproperty
    def registered(self) -> bool:
        """All providers have been registered"""
        pass

    @abstractproperty
    def booted(self) -> bool:
        """All providers have been booted"""
        pass

    @abstractproperty
    def is_console(self) -> bool:
        """App running from CLI (not serving web or API)"""
        pass

    @abstractproperty
    def is_http(self) -> bool:
        """App running as HTTP server (not as CLI)"""
        pass

    @abstractproperty
    def packages(self) -> Dict[str, Package]:
        """OrderedDict of all packages defined from providers"""
        pass

    @abstractproperty
    def path(self) -> str:
        """Base path of running application"""
        pass

    @abstractproperty
    def name(self) -> str:
        """Short name of running application"""
        pass

    @abstractproperty
    def main(self) -> str:
        """The main packages running this application"""
        pass

    @abstractmethod
    def bootstrap(self, app_config: Dict, path: str, is_console: bool) -> None:
        """Bootstrap the uvicore application"""
        pass

    @abstractmethod
    def package(self, package: str, *, name: str = None, main: bool = False) -> Package:
        """Get package by name or by main running package"""
        pass

    @abstractmethod
    def perf(self, item: any) -> None:
        """Add entry to debug performance counter"""
        pass

    # @abstractmethod
    # def dump(self, *args) -> None:
    #     """Pretty print args to console"""
    #     pass

    # @abstractmethod
    # def dd(self, *args) -> None:
    #     """Pretty print args to console and exit()"""
    #     pass



import uvicore
from uvicore.typing import Dict, Callable, ASGIApp, Send, Receive, Scope
from uvicore.http.response import Text, HTML, JSON, Response
from uvicore.http.request import HTTPConnection, Request
from uvicore.support.dumper import dump, dd
from uvicore.support import module
from uvicore.http.exceptions import HTTPException
from uvicore.contracts import Authenticator, UserInfo, UserProvider


@uvicore.service()
class Authentication:
    """Authentication global middleware capable of multiple authenticator backends"""

    def __init__(self, app: ASGIApp, route_type: str) -> None:
        # __init__ called one time on uvicore HTTP bootstrap
        # __call__ called on every request
        self.app = app
        self.route_type = str(route_type or 'api').lower()
        assert self.route_type in ['web', 'api']

        # Load and merge the auth config for this route type (web or api)
        self.config = get_auth_config(self.route_type)

        # Variable caching
        # Middleware needs to be fast.  Really fast!
        # Even saving a few CPU cycles here and there makes a huge different
        # with a high concurrency site (or wrk benchmark).  So I use variable
        # caching to load, instantiate, clone...ONCE.  So __call__ need to be efficient.
        self.cached_authenticators = {}
        self.cached_providers = {}
        self.cached_default_provider_options = {}
        self.cached_default_provider_method = None

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        # Middleware only for http and websocket types
        if scope["type"] not in ["http", "websocket"]:
            # Next middleware in stack
            await self.app(scope, receive, send)
            return

        # Get the http connection.  This is essentially the Request but not HTTP specific.
        # The base connection that works for HTTP and WebSockets.  Request inherits from HTTPConnection
        request = HTTPConnection(scope)

        # Loop each authenticator until one is a success
        user = None
        authenticator_name = None
        for authenticator_name, authenticator in self.config.authenticators.items():
            # Load authenticator backend from cache or module.load()
            if authenticator_name in self.cached_authenticators:
                #dump('cached authenticator')
                backend = self.cached_authenticators[authenticator_name]
            else:
                try:
                    # Load authenticator and instantiate it, saving the instance to our cached_authenticators
                    # This essentially makes the instance a singleton, which is fine because it is
                    # the 'authenticate' method that is called per request.  One instance is fine.
                    backend: Authenticator = module.load(authenticator.module).object(authenticator)
                    self.cached_authenticators[authenticator_name] = backend
                except Exception as e:
                    raise Exception('Issue trying to import authenticator module defined in app.auth config - {}'.format(str(e)))

            # Call backend authenticate() method
            # Return of False means this authentication method is not being attempted, try next authenticator
            # Return of True means this authentication method was being attempted, but failed validation, skip next authenticator
            # Return of User object means a valid user was found, skip next authenticator
            #dump(backend)
            # 8867 req/sec
            user = await backend.authenticate(request)
            # 6924 req/sec logging, 8500 req/sec without logging

            # Determine if we should continue to next authenticator
            # Return of True means this authorization method was being attempted, but failed validation
            # This means we can SKIP the next authenticators as they are not being attempted.
            if isinstance(user, bool) and user == True:
                break

            # If User object returned, validation success, stop authenticator itteration
            if isinstance(user, UserInfo):
                break

            # Check for exception and return
            # Should never happen as authenticators never produce an Exception
            # But if it ever does, then I can handle it properly here
            if isinstance(user, HTTPException):
                return await self.error_response(user, scope, receive, send)

        # If valid user is logged in, append 'authenticated' to permissions
        if isinstance(user, UserInfo):
            if 'authenticated' not in user.permissions:
                user.permissions.insert(0, 'authenticated')

        # Retrieve anonymous user from default_provider backend
        # If all authenticators returned no User object, user is not logged in with any method.
        # Build an anonymous user object to inject into the request scope
        if not isinstance(user, UserInfo):
            # 8350 req/sec
            user = await self.retrieve_anonymous_user(request)
            # 6600 req/sec - even though I am now using ARRAY cache store
            # to store the user.  This must mean pickle is slowing it down.
            # 7500 without pickel!  Don't think array needs pickling because
            # its already python!

            # Ensure authenticated is False
            user.authenticated = False

            # No, will never happen, its a new user
            # if 'authenticated' in user.permissions:
            #     user.permissions.remove('authenticated')

        # Add user to request
        scope['user'] = user
        #scope["auth"] = user.permissions  # No, this was starlette example of where to place scopes

        # Add route_type (web or api) to request for guard.py usage
        scope['route_type'] = self.route_type

        # Add matched authenticator to request for later usage (like logout functionality)
        scope['authenticator'] = authenticator_name

        # Next global middleware in stack
        await self.app(scope, receive, send)

    async def retrieve_anonymous_user(self, request: HTTPConnection):
        """Retrieve anonymous user from User Provider backend"""
        # Import user provider defined in auth config
        provider_module = self.config.default_provider.module

        if provider_module in self.cached_providers:
            #dump('cached provider')
            user_provider = self.cached_providers[provider_module]
        else:
            #dump('UNcached provider')
            # Anonymous user provider is always the 'default_provider'
            # Load and instantiate just once, then cache the results, a huge boost for middleware!
            user_provider: UserProvider = module.load(self.config.default_provider.module).object()
            self.cached_providers[provider_module] = user_provider

        # Get additional provider kwargs from options and anonymous_options config
        if not self.cached_default_provider_options:
            # Cached because of the .clone(), this is a nice boost for middleware!
            self.cached_default_provider_options = self.config.default_provider.options.clone()
            self.cached_default_provider_options.merge(self.config.default_provider.anonymous_options)  # Anonymous wins in merge

        # Alter provider method called based on existence of username, id, or uuid parameters
        if not self.cached_default_provider_method:
            # No need to perform an 'in' statement on each request.  The first request
            # simply cache the default providers retrieval method
            self.cached_default_provider_method = user_provider.retrieve_by_username
            if 'id' in self.cached_default_provider_options: self.cached_default_provider_method = user_provider.retrieve_by_id
            if 'uuid' in self.cached_default_provider_options: self.cached_default_provider_method = user_provider.retrieve_by_uuid

        # Call user provider method passing in defined self.cached_default_provider_options
        # 57600 req/sec
        return await self.cached_default_provider_method(request=request, **self.cached_default_provider_options)
        # 51489. req/sec
        # Wow, this call to retrieve user is a bit slow, even with 'array' cache forced without pickle.

    async def error_response(self, user: UserInfo, scope: Scope, receive: Receive, send: Send):
        """Build and return error response"""
        if self.route_type == 'web':
            # Fixme, how can I use the user customized 401 HTML?
            response = HTML(
                content='401 HTML here? - {}'.format(user.detail),
                status_code=user.status_code,
                headers=user.headers
            )
        else:
            response = JSON(
                content={'message': user.detail},
                status_code=user.status_code,
                headers=user.headers
            )
        if scope["type"] == "websocket":
            await send({"type": "websocket.close", "code": 1000})
        else:
            await response(scope, receive, send)
        return



# Stand-alone function because we use this elsewhere outside the Authentication class
def get_auth_config(route_type: str = 'aip'):
    """Get web or api auth config and merge default options and providers"""

    # Load all default option configs
    default_options = uvicore.config.app.auth.default_options

    # Load all provider configs
    providers = uvicore.config.app.auth.providers

    # Merge default options and providers into each authenticator
    config_path = 'app.auth.api'
    if route_type == 'web': config_path = 'app.auth.web'
    config = uvicore.config.dotget(config_path).clone()

    # Merge provider Dict if specified as a string
    if 'default_provider' in config and isinstance(config.default_provider, str):
        if config.default_provider in providers:
            config.default_provider = providers[config.default_provider].clone()
        else:
            raise Exception("Default options '{}' not found in auth config".format(config.default_provider))

    # Merge each providers configuration
    for authenticator in config.authenticators.values():

        # Merge default_options Dict if specified as a string
        if 'default_options' in authenticator:
            # Deep merge default options
            if authenticator.default_options in default_options:
                authenticator.defaults(default_options[authenticator.default_options])  # Defaults does a clone!
            else:
                raise Exception("Default options '{}' not found in auth config".format(authenticator.default_options))

        # If no provider specified, use default_provider (already a full Dict from above)
        if 'provider' not in authenticator:
            authenticator.provider = config.default_provider

        # Merge provider Dict if specified as a string
        if 'provider' in authenticator and isinstance(authenticator.provider, str):
            if authenticator.provider in providers:
                authenticator.provider = providers[authenticator.provider].clone()
            else:
                raise Exception("Provider '{}' not found in auth config".format(authenticator.provider))

    # Returned merge config for all authenticators
    return config




















# import typing

# from starlette.authentication import (
#     AuthCredentials,
#     AuthenticationBackend,
#     AuthenticationError,
#     UnauthenticatedUser,
# )
# from starlette.requests import HTTPConnection
# from starlette.responses import PlainTextResponse, Response
# from starlette.types import ASGIApp, Receive, Scope, Send


# class AuthenticationMiddleware:
#     def __init__(
#         self,
#         app: ASGIApp,
#         backend: AuthenticationBackend,
#         on_error: typing.Callable[
#             [HTTPConnection, AuthenticationError], Response
#         ] = None,
#     ) -> None:
#         self.app = app
#         self.backend = backend
#         self.on_error = (
#             on_error if on_error is not None else self.default_on_error
#         )  # type: typing.Callable[[HTTPConnection, AuthenticationError], Response]

#     async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
#         if scope["type"] not in ["http", "websocket"]:
#             await self.app(scope, receive, send)
#             return

#         conn = HTTPConnection(scope)
#         try:
#             auth_result = await self.backend.authenticate(conn)
#         except AuthenticationError as exc:
#             response = self.on_error(conn, exc)
#             if scope["type"] == "websocket":
#                 await send({"type": "websocket.close", "code": 1000})
#             else:
#                 await response(scope, receive, send)
#             return

#         if auth_result is None:
#             auth_result = AuthCredentials(), UnauthenticatedUser()
#         scope["auth"], scope["user"] = auth_result
#         await self.app(scope, receive, send)

#     @staticmethod
#     def default_on_error(conn: HTTPConnection, exc: Exception) -> Response:
#         return PlainTextResponse(str(exc), status_code=400)



import uvicore
from uvicore.support.dumper import dd, dump
from uvicore.contracts import Config as ConfigInterface


@uvicore.service('uvicore.configuration.configuration.Configuration',
    aliases=['Configuration', 'Config', 'configuration', 'config'],
    singleton=True,
)
class Configuration(ConfigInterface):
    pass










# OBSOLETE, now SuperDict!

import os
from typing import Any, Dict
from uvicore.support.dictionary import deep_merge
class _ConfigurationOBSOLETE(ConfigInterface):
    """Configuration private class.

    Do not import from this location.
    Use the uvicore.config singleton global instead."""

    #config: Dict = {}
    @property
    def items(self) -> Dict[str, Dict]:
        return self._items

    def __init__(self) -> None:
        self._items: Dict = {}

    def __call__(self, dotkey: str = None):
        return self.get(dotkey)

    def get(self, dotkey: str = None, _recursive_config: Dict = None) -> Any:
        # Recursive for dot notation
        if not dotkey:
            return self.items
        if _recursive_config is None: _recursive_config = self.items
        if "." in dotkey:
            key, rest = dotkey.split(".", 1)
            if key not in _recursive_config:
                _recursive_config[key] = {}
            return self.get(rest, _recursive_config[key])
        else:
            if dotkey in _recursive_config:
                return _recursive_config[dotkey]
            else:
                return None

    def set(self, dotkey: str, value: any, _recursive_config: Dict = None) -> Any:
        # Recursive for dot notation
        # Remember objects are byRef, so changing config also changes self.items
        if _recursive_config is None: _recursive_config = self._items
        if "." in dotkey:
            key, rest = dotkey.split(".", 1)
            if key not in _recursive_config:
                _recursive_config[key] = {}
            return self.set(rest, value, _recursive_config[key])
        else:
            _recursive_config[dotkey] = value
            return value

    def merge(self, dotkey: str, value: Any) -> None:
        existing = self.get(dotkey);
        if not existing:
            return self.set(dotkey, value)
        else:
            # Perform deep merge on existing config
            # This overrides 'existing' data with 'value' data
            # And updates 'existing' with merged vlues
            existing = deep_merge(value, existing)

            # Finally set the new merged config
            self.set(dotkey, existing)


# IoC Class Instance
# Not to be imported by the public from here.
# Use the uvicore.config singleton global instead.

# Public API for import * and doc gens
#__all__ = ['_Configuration']



import uvicore
from uvicore.typing import Callable, Decorator, Any, Dict, List
from uvicore.support.dumper import dump, dd


def service(name: str = None, *, factory: Any = None, kwargs: Dict = None, singleton: bool = False, aliases: List = []) -> Callable[[Decorator], Decorator]:
    def decorator(cls: Decorator) -> Decorator:
        # Bind this service into the Ioc
        bind_name = name or cls.__module__ + '.' + cls.__name__
        new_cls = uvicore.ioc.bind_from_decorator(cls, name=bind_name, object_type='service', factory=factory, kwargs=kwargs, singleton=singleton, aliases=aliases)

        # Other things with a table

        return new_cls
    return decorator



import uvicore
from uvicore.auth.support import password
from uvicore.auth.models.role import Role
from uvicore.auth.models.group import Group
from uvicore.support.dumper import dump, dd


@uvicore.seeder()
async def seed():
    uvicore.log.item('Seeding table groups')

    # Get all roles
    roles = await Role.query().key_by('name').get()

    await Group.insert_with_relations([
        {
            'name': 'Administrator',
            'roles': [
                roles['Administrator']
            ]
        },
        {
            'name': 'User',
            'roles': [
                roles['User']
            ]
        },
    ])

    #await Group(name='Administrator').save()
    #await Group(name='Users').save()




import uvicore
from uvicore.typing import Dict, Tuple
from uvicore.http import Request
from fastapi.security import SecurityScopes
from uvicore.support.dumper import dump, dd
from uvicore.http.exceptions import HTTPException, PermissionDenied, InvalidCredentials
from uvicore.http import status
from uvicore.support import module
import jwt
from uvicore.auth.middleware.auth import Auth


@uvicore.service()
class Jwt(Auth):
    """Jwt Authentication Route Middleware"""

    def __init__(self, config: Dict):
        self.config = config

    async def __call__(self, scopes: SecurityScopes, request: Request):
        dump('JWT HERE')
        dump(scopes.__dict__, self.config)
        dump('-----------------------------------------------------------------')
        # Assume unauthorized
        authorized = False

        # Parse authorization header
        authorization, scheme, token = self.auth_header(request)

        #dump("AUTHORIZATION:", authorization)
        #dump("SCHEME", scheme)
        #dump("TOKEN", token)
        dump('HEADERS', request.headers)

        # Detect anonymous API access.
        # API gateways that allow anonymous access to an API will pass the request
        # downstream even if the token is missing, invalid or expired.
        # When they do, they add a header notifying downstream that auth has failed
        # but anonymous access is still allowed.  For Kong this header is x-anonymous-consumer
        # If the API gateway does not allow anonymous access, downstream won't even be proxied.
        if (self.config.api_gateway.enabled
            and self.config.api_gateway.anonymous_header
            and self.config.api_gateway.anonymous_header in request.headers
        ):
            # Anonymous header configured and found in headers.  This is an anonymous connection.
            # No need to check JWT, simply return anonymous "public" user.
            dump('ANONYMOUS CONNECTION, RETURN PUBLIC USER')
            return




        # Kong with NO JWT using anonymous example
        # headers
        """
        {'host': 'sunjaro:5000', 'connection': 'keep-alive',
        'x-forwarded-for': '127.0.0.1, 172.26.0.1', 'x-forwarded-proto': 'https',
        'x-forwarded-host': 'uvicore-local.sunfinity.io',
        'x-forwarded-port': '8443', 'x-real-ip': '172.26.0.1',
        'user-agent': 'HTTPie/2.4.0', 'accept-encoding': 'gzip, deflate',
        'accept': '*/*',

        # This is anonymous kong ID if x-anonymous-consumer = True
        'x-consumer-id': '1be13da8-296d-4653-bb68-cdc7f4b54a6d',
        'x-consumer-custom-id': 'anonymous',
        'x-consumer-username': 'anonymous',
        'x-anonymous-consumer': 'true'})
        """


        # Headers, Kong with valid token
        """
        {'host': 'sunjaro:5000', 'connection': 'keep-alive',
        'x-forwarded-for': '127.0.0.1, 172.26.0.1', 'x-forwarded-proto': 'https',
        'x-forwarded-host': 'uvicore-local.sunfinity.io', 'x-forwarded-port': '8443',
        'x-real-ip': '172.26.0.1', 'user-agent': 'HTTPie/2.4.0', 'accept-encoding': 'gzip, deflate',
        'accept': '*/*',
        'authorization': 'Bearer eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCIsImtpZCI6Ii1GV1Z1eEg0eGh2NnhxQUV5akVfdXBfd0NiZyJ9.eyJhdWQiOiJkNzA5YTQzMi01ZWRjLTQ0YzctODcyMS00Y2YxMjM0NzNjNDUiLCJleHAiOjE2MTUzMzE0OTAsImlhdCI6MTYxNTMyNzg5MCwiaXNzIjoiaHR0cHM6Ly9hdXRoLWxvY2FsLnN1bmZpbml0eS5jb20iLCJzdWIiOiIxNGEzZjY0Yi0zNzcwLTQ1NDktOWZjNS00YjYxZmY4YjIwNzAiLCJhdXRoZW50aWNhdGlvblR5cGUiOiJQQVNTV09SRCIsImVtYWlsIjoibWFuYWdlcjFAZXhhbXBsZS5jb20iLCJlbWFpbF92ZXJpZmllZCI6dHJ1ZSwiYXBwbGljYXRpb25JZCI6ImQ3MDlhNDMyLTVlZGMtNDRjNy04NzIxLTRjZjEyMzQ3M2M0NSIsInJvbGVzIjpbIkVtcGxveWVlIiwiU2FsZXMgSW50ZXJuYWwiLCJVc2VyIl0sIm5hbWUiOiJNYW5hZ2VyfE9uZSJ9.JaR8EdMdd5XdUBZuPWXm-uVghODp84LzQON3kwwshqrjR5PyAO6HP6S5wGg1Y-dKMfETkbNy5sh9RP3nZWHtwB7ZukQ3IwIrC5Ako3a1VZWFzSu1mR3_1X510DvTQVwdfWz_XvEwG5v8O01n-UrLXWQHvgKDeEs3HPkuel-kHXajhoP6oEVxXqb3AmzCIoOv9MC3_3F4jVW73FLFQA-thsVElpJiwrLlZ9ehulp-Cl92eCOS6Ug-L6aJdafx7X9Q2gTsY28cN85WVCX3dss3gJ1GO8vq8u8yIek4B-27hRtyVwaGlxKWjGPsLjE-M0VvXSqQrCIszvxlcodo1L1F5A',

        # This is kongs consumer GUID
        'x-consumer-id': '304ae0c3-de56-47dc-8c35-7c8ed7f6c08f',

        # This is the uvicore app ID
        'x-consumer-custom-id': 'd709a432-5edc-44c7-8721-4cf123473c45',
        'x-consumer-username': 'Uvicore',
        'x-credential-identifier': 'd709a432-5edc-44c7-8721-4cf123473c45'})
        """




        # This authorization method not provided or attempted, goto next guard in middleware stack
        if not authorization or scheme != "bearer":
            # Return None means goto next authenticator in authorization middleware
            return None

        # Decode JWT
        try:
            if self.config.verify_signature:
                # With secret algorithm validation
                dump('WITH Validation')
                payload = Dict(jwt.decode(token, self.config.secret, audience=self.config.audience, algorithms=self.config.algorithms))
            else:
                # Without validation
                dump('WITHOUT Validation')
                payload = Dict(jwt.decode(token, options={"verify_signature": False}))

                # Validate aud claim
                # jwt.decode also validates the aud claim.  Since we are skipping validation, we'll still validate aud claim here.
                if payload.aud != self.config.audience:
                    raise HTTPException(
                        status_code=status.HTTP_401_UNAUTHORIZED,
                        detail="Invalid audience",
                    )
        except Exception as e:
            # Issue with validation.  Bad key, token expired...
            # Pass JWT library exception message right through with a generic 401
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail=str(e),
            )

        dump('JWT', payload)

        # {
        #     'aud': '222b06eb-85ce-472b-af30-ec09244e3bf0',
        #     'exp': 1614980706,
        #     'iat': 1614977106,
        #     'iss': 'https://auth-local.sunfinity.com',
        #     'sub': '217e27b4-0e0a-464c-84a8-c40312b55801',
        #     'authenticationType': 'PASSWORD',
        #     'email': 'it@sunfinity.com',
        #     'email_verified': True,
        #     'applicationId': '222b06eb-85ce-472b-af30-ec09244e3bf0',
        #     'roles': ['Administrator'],
        #     'name': 'Admin|Istrator'
        # }

        # Get user and validate credentials
        user = await self.retrieve_user(payload.email, None, self.config.provider)

        # If user is none and auto_create_user is enabled, auto-create user
        # Link user up to groups table based on JWT roles
        # Now self.get_user again

        # If no user returned, validation has failed or user not found
        if user is None: raise InvalidCredentials()

        # Validate Permissions
        #self.validate_permissions(user, scopes.scopes)

        # Authorization successful.
        # Add user to request in case we use it in a decorator, we can pull it out with request.scope.get('user')
        request.scope['user'] = user

        # Return user in case we are using this guard as a dependency injected route parameter
        return user




from uvicore.support.dumper import dump, dd
from uvicore.support.module import location
from uvicore.database import Connection
from uvicore.typing import Dict, List


class Db:
    """Database Service Provider Mixin"""

    #def _add_db_definition(self, key, value):
        # if type(value) == list:
        #     if not self.package.database[key]: self.package.database[key] = []
        #     self.package.database[key].extend(value)
        # elif type(value) == dict:
        #     self.package.database[key].merge(value)
        # else:
        #     self.package.database[key] = value

        #self.package.database[key] = value

        # if 'database' not in self.package:
        #     self.package.database = Dict()
        # if type(value) == list:
        #     if key not in self.package.database: self.package.database = []
        #     self.package['database'][key].extend(value)
        # else:
        #     self.package['database'][key] = value

    def register_db_connections(self, connections: Dict, default: str):
        #connections = []
        for name, connection in connections.items():

            # Build URL and metakey
            # Metakey cannot be the connection name.  If 2 connections share the exact
            # same database (host, port, dbname) then they need to also share the same
            # metedata for foreign keys to work properly.
            if not connection.backend: connection.backend = 'sqlalchemy'
            connection.backend = connection.backend.lower()
            connection.driver = connection.driver.lower()
            url = ''
            if connection.backend == 'sqlalchemy':
                if connection.driver == 'sqlite':
                    url = 'sqlite:///' + connection.database
                    metakey = url
                elif connection.driver in ['mysql', 'postgresql']:
                    url = connection.driver
                    if connection.dialect: url += '+' + connection.dialect
                    url += (
                        '://' + connection.username
                        + ':' + connection.password
                        + '@' + connection.host
                        + ':' + str(connection.port)
                        + '/' + connection.database
                    )
                    metakey = (
                        connection.host
                        + ':' + str(connection.port)
                        + '/' + connection.database
                    )
            else:
                # Any opther backend type (like api), should have at minimum a url defined
                url = connection.url
                metakey = url


            # Merge new values into connection SuperDict
            if not connection.prefix: connection.prefix = ''
            connection.merge({
                'name': name,
                'metakey': metakey.lower(),
                'url': url
            })

        self.package.database.connections = connections
        self.package.database.connection_default = default

    def register_db_models(self, items: List):
        # Default registration
        self.package.registers.defaults({'models': True})

        # Register models only if allowed
        if self.package.registers.models:
            self.package.database.models = items

    def register_db_tables(self, items: List):
        # Default registration
        self.package.registers.defaults({'tables': True})

        # Register tables only if allowed
        if self.package.registers.tables:
            self.package.database.tables = items

    def register_db_seeders(self, items: List):
        # Default registration
        self.package.registers.defaults({'seeders': True})

        # Register seeders only if allowed
        if self.package.registers.seeders:
            self.package.database.seeders = items



import uvicore
from uvicore.console import command, argument, option
from uvicore.support.dumper import dd, dump


@command()
def info():
    """Running Application Information"""
    uvicore.log.header("Final Merged Application Information")
    uvicore.log.line()

    dump(uvicore.app._running_config)




from uvicore.typing import Any, Dict, Tuple, List
from collections import OrderedDict
from collections import OrderedDict
from uvicore.support.dictionary import deep_merge
from uvicore.support.dumper import dump, dd


def haskey(object: Any, key: Any) -> bool:
    """Check if value exists on a dict or class instance attribute value in a unified way"""
    if type(object) == dict or type(object) == OrderedDict:
        # Dict or OrderedDict
        return key in object
    else:
        # Class instance
        return hasattr(object, key)


def getvalue(object: Any, key: Any) -> Any:
    """Access a dict or class instance attribute value in a unified way"""
    if type(object) == dict or type(object) == OrderedDict:
        # Dict or OrderedDict
        return object.get(key)
    else:
        # Class instance
        if hasattr(object, key):
            return getattr(object, key)
        else:
            return None


def getitems(object: Any) -> List[Tuple]:
    """Get all key/value from a dict or class instance for consistent .items() iteration"""
    # With a dict you can itterate key/value with for (key,value) in dict.items()
    # but with a class you cannot.  This standardizes the items() for Dict or class.
    if type(object) == dict:
        # Dict or OrderedDict
        return object.items()
    else:
        # Class instance
        fields = []
        if getattr(object, '__fields__'):
            # Pydantic model
            fields = object.__fields__
        else:
            # Python class
            # vars(object) is the same as object.__dict__
            fields = [x for x in vars(object) if not x.startswith('__') and not callable(getattr(object, x))]
        # Return as List[Tuple] [(key, value)]
        return [(field, getattr(object, field)) for field in fields]


def setvalue(object: Any, key: Any, value: Any) -> None:
    """Set a dict or class instance attribute value in a unified way"""
    if type(object) == dict or type(object) == OrderedDict:
        # Dict or OrderedDict
        object[key] = value
    else:
        # Class instance

        setattr(object, key, value)


def dotget(object: Dict, dotpath: str, default = None):
    """Access a dict or class instance attribute value by dot notation which handles nested null values well"""

    # Eliminates the need to do this if you think a child attribute may not exist
    # BAD  connection_default=custom_config.get('database').get('default') if 'database' in custom_config else None,
    # GOOD connection_default=dotget(custom_config, 'database.default')

    if '.' in dotpath:
        paths = dotpath.split('.')
    else:
        paths = [dotpath]

    node = object
    for path in paths:
        node = getvalue(node, path)
        if node is None: return default
    return node


def unique(object: List):
    """Remove duplicates from a list while preserving the order"""
    # list(set(object)) does the unique job, but it sets to random order
    u = []
    for item in object:
        if item not in u:
            u.append(item)
    return u




### Below is junk, experimental
# These once actually used, would go into uvicore.types instead

class Str:
    def __init__(self, data):
        self.__data = data

    def contains(self, value):
        return value in self.__data

    def append(self, value):
        return str(self.__data) + str(value)

    def upper(self):
        return str(self.__data).upper()

    def lower(self):
        return str(self.__data).lower()


class Obj:
    def __init__(self, data):
        self.__dict__ = data

    def dict(self):
        return {k: v for k, v in self.__dict__.items() if not k.startswith('__')}

    def __repr__(self):
        return str({k: v for k, v in self.__dict__.items() if not k.startswith('__')})


class Collection():
    # _names are protected
    # __names are private
    # @property for getters
    # @varname.setter for setters

    def __init__(self, data):
        #self.test = 'asdf'
        self.__i = 0
        self.__data = []
        #self.Object = lambda **kwargs: type("Object", (), kwargs)
        for item in data:
            #self.__data.append(self.__object(item))
            self.__data.append(Obj(item))

    def dict(self):
        items = []
        for data in self.__data:
            items.append(data.dict())
        return items

    def filter(self, callback):
        return filter(callback, self.__data)

    # def __object(self, dict):
    #     return type("Object", (), dict)

    def add(self, item):
        #self.__data.append(self.__object(item))
        self.__data.append(Obj(item))

    def __iter__(self):
        return iter(self.__data)

    def __getitem__(self, item):
        return self.__data[item]

    # @property
    # def test(self):
    #     return self.__test

    # @test.setter
    # def test(self, value):
    #     self.__test = value

    # def __next__(self):
    #     #return self.next()
    #     print('1')
    #     if len(self.data) <= self.i:
    #         self.i += 1
    #         #return self.data[self.i]
    #         return self.i
    #     else:
    #         raise StopIteration()




import json
from uvicore.typing import Optional, Any, Dict
from starlette.responses import HTMLResponse
from fastapi.openapi.docs import get_swagger_ui_oauth2_redirect_html
from fastapi.encoders import jsonable_encoder

def get_rapidoc_ui_html(
    *,
    openapi_url: str
) -> HTMLResponse:

    html = f"""
        <!doctype html>
        <head>
        <script type="module" src="https://unpkg.com/rapidoc/dist/rapidoc-min.js"></script>
        </head>
        <body>
        <rapi-doc spec-url="{openapi_url}" >
        </rapi-doc>
        </body>
    """
    return HTMLResponse(html)


def get_rapidoc_pdf_ui_html(
    *,
    openapi_url: str
) -> HTMLResponse:
    html = f"""
        <!doctype html>
        <html>
        <head>
        <script src="https://unpkg.com/rapipdf/dist/rapipdf-min.js"></script>
        </head>
        <body>
        <rapi-pdf
            style = "width:700px; height:40px; font-size:18px;"
            spec-url = "{openapi_url}"
            button-bg = "#b44646"
        > </rapi-pdf>
        </body>
        </html>
    """
    return HTMLResponse(html)


def get_rapidoc_ui_oauth2_redirect_html() -> HTMLResponse:
    html = f"""
        <!DOCTYPE html>
        <html lang=en>
        <head>
            <title>RapiDoc UI</title>
            <meta charset="utf-8">
            <meta name="viewport" content="width=device-width,minimum-scale=1,initial-scale=1,user-scalable=yes">
            <script type="module" src="rapidoc-min.js"></script>
        </head>
        <body>
            <noscript><strong>We're sorry but RapiDoc doesn't work properly without JavaScript enabled. Please enable it to continue.</strong></noscript>
            <oauth-receiver> </oauth-receiver>
        </body>
        </html>
    """
    return HTMLResponse(html)



import re


def ucbreakup(value: str):
    """Breakup words with by - _ or camel and uppercase first letter of each word stripping non alphanumeric"""
    #dashes = re.compile('_|-')
    delimiter = '_'
    uppers = re.compile(r'(?<!^)(?=[A-Z])')
    nonalphanumeric = re.compile('[^0-9a-zA-Z\ _-]+')
    dashes = re.compile('\ |_|-')

    #value = dashes.sub(delimiter, value)
    value = uppers.sub(' ', value)
    value = nonalphanumeric.sub('', value)

    value = dashes.sub(' ', value)
    value = ucwords(value)
    #value = value.replace(' ', '')
    #value = uppers.sub(' ', value)
    return value


def ucwords(value: str):
    """Uppercase first letter of each word but leaves rest of the word alone (keeps existing case)"""
    words = []
    for word in value.split(' '):
        if word: words.append(word[0].upper() + word[1:])
    return ' '.join(words)


def ucfirst(value: str):
    """Uppercase first letter in sentence and leaves rest alone"""
    return value[0].upper() + value[1:]


def title(value: str):
    """Uppsercase first letter of each word and forces lowercase on the rest of the word"""
    return value.title()


def snake(value: str, delimiter: str = '_'):
    """Convert string to snake case foo_bar"""
    return slug(value, '_')


def kebab(value: str):
    """Convert string to kebab case foo-bar"""
    #return snake(value, '-')
    return slug(value, '-')


def studly(value: str):
    """Convert string to studly/pascal case FooBar"""
    value = ucbreakup(value)
    return value.replace(' ', '')


def camel(value: str):
    """Convert string to camel case fooBar"""
    value = studly(value)
    #value = ucbreakup(value)
    value = value[0].lower() + value[1:]
    return value


def slug(value: str, delimiter: str = '-'):
    """Slugify a string"""
    value = ucbreakup(value)
    return value.replace(' ', delimiter).lower()


def words(value: str, length: int = 100, end: str = '...'):
    words = value.split(' ')
    if len(words) > length:
        return ' '.join(words[0:length]) + str(end)
    return value



from __future__ import annotations
import uvicore
import sqlalchemy as sa
from uvicore.orm.mapper import Mapper
from pydantic import main as PydanticMain
from uvicore.support.dumper import dd, dump
from uvicore.orm.query import OrmQueryBuilder
from uvicore.support.classes import hybridmethod
from uvicore.contracts import Model as ModelInterface
from uvicore.support.collection import getvalue, setvalue
from uvicore.typing import Any, Dict, Generic, List, Tuple, TypeVar, Union
from uvicore.orm.fields import BelongsTo, BelongsToMany, Field, HasMany, HasOne, MorphMany, MorphOne, MorphToMany

E = TypeVar("E")

# Monkey Patch Pydantics Main validate_field_name.  Why?
# Because if you EXTEND a model from another model, and your extension has
# fieldnames that match fieldnames in metaclass.py, you will receive the
# Field name xxx shadows a BaseModel attribute.  Only happens when you extend.  Normally
# a field and a metaclass.py property will not collide.  Extension will happen often, especially
# extending the main uvicore.auth User model to add custom fields to user table.
def validate_field_name(bases, field_name) -> None:
    # mReschke modify from original method in pydantic utils.py.
    # Only check for field clashes if we are not extending another model
    for base in bases:
        # Can't check isinstance because even if extended it will always be true
        if base.__class__.__module__ == 'pydantic.main':
            # We are not extending another Uvicore model.  If we were it class module would be uvicore.orm.model
            if getattr(base, field_name, None):
                raise NameError(
                    f'Field name "{field_name}" shadows a BaseModel attribute; '
                    f'use a different field name with "alias=\'{field_name}\'".'
                )
PydanticMain.validate_field_name = validate_field_name
PydanticBaseModel = PydanticMain.BaseModel


# Any method here in the model WILL CLASH with your models field names but do show up in code intellisense.
# However any methods you add to the meteclass will NOT clash with fieldnames (even if you
# extend because of the validate_field_names monkey patch above), but metaclass methods will
# NOT show up in an interface or code intellisense.
# So reserved field names are:
#   query
#   insert
#   insert_with_relations
#   mapper
#   create
#   save
#   delete
#   link
#   unlink
# and other items inside pydantic BaseModel (main.py) that will error itself like:
#   dict
#   json
#   parse_obj


@uvicore.service()
class Model(Generic[E], PydanticBaseModel, ModelInterface[E]):

    def __init__(self, **data: Any) -> None:
        # Call pydantic parent
        super().__init__(**data)

        # Fill in callback properties
        for (key, callback) in self.__class__.__callbacks__.items():
            setattr(self, key, callback(self))

    @classmethod
    def query(entity) -> OrmQueryBuilder[OrmQueryBuilder, E]:
        return OrmQueryBuilder(entity)

    # These are nice, but they polute the namespace of FIELDS, so use just query()
    # @classmethod
    # async def get(entity) -> List[Any]:
    #     return await OrmQueryBuilder(entity).get()

    # @classmethod
    # async def find(entity, id: Any) -> Any:
    #     return await OrmQueryBuilder(entity).find(id)

    # @classmethod
    # def include(entity, *args):
    #     return OrmQueryBuilder(entity).include(*args)

    @classmethod
    async def insert(entity, models: Union[E, Dict, List[E], List[Dict]]) -> Any:
        """Insert one or more entities as List of entities or List of Dictionaries

        This bulk insert does NOT allow inserting child relations at the
        same time as there is no way to get each parents PK out to
        reference with each child in BULK. If you want to insert parent
        and relations at the same time use the slower non-bulk
        insert_with_relations() instead.
        """

        # Convert any type of dict or list to an actual Model or List[Model]
        models = entity.mapper(models).model(perform_mapping=False)

        # There is no way to check if record already exists because these new models
        # have no primary key yet.  If you run it twice with the same data, it will just
        # insert twice.  So we are assuming these are all new records therefore we CAN
        # fire the _before_insert hooks

        # Loop each model and call the before_save hook
        if type(models) == list:
            for model in models:
                # Fire model hooks
                await model._before_insert()
                await model._before_save()
        else:
            await models._before_insert()
            await models._before_save()

        # Convert List[Model] or List[Dict] into dict of mapped table columns ready for insert
        bulk = entity.mapper(models).table()
        query = entity.table.insert()

        result = None
        if type(bulk) == list:
            # List, so bulk insert
            result = await entity.execute(query, bulk)
        else:
            # Single, so single insert, returning PK or silently passing if error
            # WHY? Was I silently passing?
            #try:
            result = await entity.execute(query, bulk)
            #except:
                #pass

        # Loop each model and call the after_save hook
        if type(models) == list:
            for model in models:
                await model._after_insert()
                await model._after_save()
        else:
            await models._after_insert()
            await models._after_save()

        # Return insert results (if single will be PK)
        return result

    @classmethod
    async def insert_with_relations(entity, models: List[Dict], *, parent_pk = None, skip_save: bool = False) -> None:
        """Insert one or more entities as List of Dict that DO have relations included

        Because relations are included, this insert is NOT bulk and must
        loop each row, insert the parent, get the PK, then insert each
        children (or children first then parent depending on BelongsTo vs
        HasOne or HasMany)
        """

        # Note about bulk insert with nested relations Dictionary
        # Bulk insert does not give back the primary keys for each inserted record
        # So I have no way to know the PK of the parent when I insert the child relations
        # So the only way to possibly insert parent and child is to do one at a time
        # which is not bulk and will be slower.  If I wanted to do this I could pull
        # out all models that HAVE a relation and set those asside.  All that do NOT
        # have a relation I could bulk insert, then after insert one those that do have a
        # relation one at a time.  The problem with this is they won't be inserted in the
        # order you wanted.  Only way to preserve order is to check if ANY models have
        # a relation, if so, loop and insert one at a time, no bulk period.
        # SQLAlchemy does have a bulk_save_objects that does return the PKs but
        # encode/databases does not impliment this.  Instead their insert_many
        # uses a cursor where each row is added as a query, then the cursor is committed.
        # This is bulk insert, but no way to retrieve each PK from it.

        # Ensure models is a list
        if type(models) != list: models = [models]

        for model in models:

            # Skip empty models (None or [])
            if not model: continue

            # If model is not a dict, its probably a real pydantic model instance, convert it to a dict
            if type(model) != dict: model = model.dict()

            # Check each field for relations and rename them before inserting parent
            relations = {}
            skip_children = False
            for (fieldname, value) in model.items():
                field = entity.modelfields.get(fieldname)  # Using direct dictionary to skip bad values in dict
                if not field: continue
                if not field.relation: continue
                relation = field.relation.fill(field)

                # Cannot change a dict in place or you get error
                # RuntimeError: dictionary keys changed during iteration
                # Se we'll add to a list and delete it outside the loop
                relations[fieldname] = {
                    'relation': relation,
                    'data': value
                }

            # Delete relations from parent model as they cannot be inserted
            for relation in relations.keys():
                del model[relation]

            # BelongTo relations are the inverse of HasOne or HasMany in that
            # the child has to be created BEFORE the parent
            for relation in relations.values():
                data = relation['data']
                relation = relation['relation']

                if type(relation) == BelongsTo:
                    # Because we insert the children first, skip inserting
                    # children later down the method
                    skip_children = True

                    # Ignore empty relation (None or []), AFTER skip_children is set
                    if not data: continue

                    # Check if one ONE relation is already a fully inserted object (presense of pk)
                    # If NOT, it needs to be inserted, if so, just grab its existing PK as the child_pk
                    if getvalue(data, relation.foreign_key):
                        # Already inserted realtion, use its existing PK
                        child_pk = getvalue(data, relation.foreign_key)
                    else:
                        # Recursively insert child and get PK
                        child_pk = await relation.entity.insert_with_relations([data])

                    # Replace parent foreignKey with child_pk
                    model[relation.local_key] = child_pk

            # Convert model Dict into actual Model instance
            model_instance = entity.mapper(model).model(perform_mapping=False)

            # Insert the parent model and retrieve parents new PK value
            # Unless we are told to skip, as is the case with ManyToMany where we
            # use .create() instead of .save()
            if not skip_save:
                model_instance = await model_instance.save()
                parent_pk = getattr(model_instance, entity.pk)

            # Now insert each child relation
            for relation in relations.values():
                data = relation['data']
                relation = relation['relation']

                # Ignore empty relation (None or [])
                if not data: continue

                childmodels = data
                if type(childmodels) != list: childmodels = [childmodels]

                # Insert BelongsTo (which is the only one that obeys skip_children)
                if type(relation) == BelongsTo and not skip_children:
                    await relation.entity.insert_with_relations(childmodels)

                # Insert HasMany or HasOne (which are basically the same)
                elif type(relation) == HasOne or type(relation) == HasMany:
                    # Replace child foreignKey with parent_pk
                    for childmodel in childmodels:
                        childmodel[relation.foreign_key] = parent_pk
                    await relation.entity.insert_with_relations(childmodels)

                # Insert Polymorphic OneToOne
                elif type(relation) == MorphOne or type(relation) == MorphMany:
                    for childmodel in childmodels:
                        childmodel[relation.foreign_type] = entity.tablename
                        childmodel[relation.foreign_key] = parent_pk
                    await relation.entity.insert_with_relations(childmodels)

                # Insert ManyToMany or Polymorphic MorphToMany
                elif type(relation) == BelongsToMany or type(relation) == MorphToMany:
                    # Insert this parent ManyToMany to get the parents PK
                    # By using .create, we are both linking the records in the pivot/relation table
                    # and also creating the actual record if it does not exist
                    await model_instance.create(relation.name, data)

                    # The data variable is byRef and is modified with the new inserted PK
                    # It will be the same for all children, so grab[0]
                    poly_parent_pk = getvalue(data[0], relation.entity.pk)

                    # Now insert any children using this parents PK, so not let it .save() so we
                    # can instead use the .create() for proper linkage
                    await relation.entity.insert_with_relations(childmodels, skip_save=True, parent_pk=poly_parent_pk)

        # Return parent_pk for recursion only
        # This method not meant to return anything valuable to the user, which
        # is why the method itself is -> None
        return parent_pk

    @hybridmethod
    def mapper(self_or_entity, *args) -> Mapper:
        """Entity mapper for model->table or table->model conversions

        Can be accessed both from the [static] class or from an instance
        """
        return Mapper(self_or_entity, *args)

    async def create(self, relation_name: str, models: Union[Any, List[Any]]) -> None:
        """Create related child records and link them to this parent (self) model"""

        # Ignore empty models (None or [])
        if not models: return

        # Get the entity of this model instance (which is the metaclass, aka self.__class__)
        entity = self.__class__

        # Get field and relation info
        field = entity.modelfield(relation_name)
        relation = field.relation.fill(field)

        # Convert to list
        if type(models) != list: models = [models]

        # Assume each does not yet exists, so create all as BULK insert
        if type(relation) == HasOne or type(relation) == HasMany or type(relation) == MorphOne or type(relation) == MorphMany:
            # Fill in relation foreign key vlaue
            for model in models:
                # Set new relation value (works if model is a dict or a model class instance!)
                setvalue(model, relation.foreign_key, getattr(self, relation.local_key))

                if type(relation) == MorphOne or type(relation) == MorphMany:
                    # For Polymorphic relations, also set type column
                    setvalue(model, relation.foreign_type, entity.tablename)

                    # # Check for Data types in values to serialize
                    # dict_values = getvalue(relation, 'dict_value')
                    # if dict_values:
                    #     if type(dict_values) != list: dict_values = [dict_values]
                    #     for dict_value in dict_values:
                    #         value = getvalue(model, dict_value)
                    #         if type(value) == dict or type(value) == list:
                    #             value = str(value)
                    #         setvalue(model, dict_value, value)

            # Bulk insert new values with proper keys
            await relation.entity.insert(models)

        # Cannot assume a record has been created or linked, so loop each and test
        # Because of this, we cannot bulk insert
        elif type(relation) == BelongsToMany or type(relation) == MorphToMany:
            for model in models:
                # If its PK is set, it already exists
                pk_value = None
                create = getvalue(model, relation.entity.pk) == None
                if create:
                    pk_value = await relation.entity.insert(model)
                    setvalue(model, relation.entity.pk, pk_value)

                # Link in pivot table
                await self.link(relation_name, model)

        else:
            raise Exception('Creating children does not work for this type of relation.')

    async def add(self, relation_name: str, models: Union[Any, List[Any]]) -> None:
        """Alias to create"""
        await self.create(relation_name, models)

    async def set(self, relation_name: str, models: Union[Any, List[Any]]) -> None:
        """Same as create(), except it deletes all first, so it sets the entire children"""

        # Get the entity of this model instance (which is the metaclass, aka self.__class__)
        entity = self.__class__

        # Get field and relation info
        field = entity.modelfield(relation_name)
        relation = field.relation.fill(field)

        if type(relation) == BelongsToMany or type(relation) == MorphToMany:
            # Delete does not work for these relations, only unlink
            await self.unlink(relation_name)
        else:
            # Delete works for these relations
            await self.delete(relation_name)
        await self.create(relation_name, models)

    async def save(self) -> Model:
        """Save this model to the database (insert or update)"""

        # Get the entity of this model instance (which is the metaclass, aka self.__class__)
        entity = self.__class__

        # Hum, think about this.  Probably a BAD idea.  Imagine you already have a record
        # and they change around relations, then type .save() would they expect ALL relations
        # to insert/delete/update properly?  I think not.  This .save() shouldn't work on relations
        # Pull out relations and try to insert/link them later
        # relations = {}
        # for fieldname in self.__dict__:
        #     field = entity.modelfields.get(fieldname)
        #     if not field: continue
        #     if not field.relation: continue
        #     data = getattr(self, fieldname)
        #     if data:
        #         relations[fieldname] = data
        #         #relation = field.relation.fill(field)
        #         # relations[fieldname] = {
        #         #     'relation': relation,
        #         #     'data': data
        #         # }

        # Check if exists
        exists = None
        table = entity.table
        if getattr(self, entity.pk):
            query = sa.select([getattr(table.c, entity.mapper(entity.pk).column())]).select_from(table).where(getattr(table.c, entity.pk) == getattr(self, entity.pk))
            #query = table.select().where(getattr(table.c, entity.pk) == getattr(self, entity.pk))  # Don't select *
            exists = await entity.fetchone(query)  # Returns None if not exists

        if exists:
            # Record exists, perform update
            await self._before_save()

            # Convert self model instance into Dict of mapped table columns, only after hooks are fired
            # only after hooks are fired as they may alter the data
            values = self.mapper().table()

            query = table.update().where(getattr(table.c, entity.pk) == getattr(self, entity.pk)).values(**values)
            await entity.execute(query)
            await self._after_save()
        else:
            # New record, perform insert
            await self._before_insert()
            await self._before_save()

            # Convert self model instance into Dict of mapped table columns
            # only after hooks are fired as they may alter the data
            values = self.mapper().table()

            query = table.insert().values(**values)
            new_pk = await entity.execute(query)

            # Only set the new_pk back to the entity if the entities PK is null
            # If not null, means its probably a string based pre-inserted PK like a 'key' field
            if getattr(self, entity.pk) is None:
                setattr(self, entity.pk, new_pk)

            await self._after_insert()
            await self._after_save()

        # This is a TRY version if if exists, not sure which is more efficient, a select first, or an insert attempt/failure
        # try:
        #     # Try insert first
        #     query = table.insert().values(**values)
        #     new_pk = await entity.execute(query)
        #     setattr(self, entity.pk, new_pk)
        # except:
        #     # If fail, record already exists
        #     query = table.update().where(getattr(table.c, entity.pk) == getattr(self, entity.pk)).values(**values)
        #     await entity.execute(query)

        # # Insert relations
        # for relation_name, data in relations.items():
        #     await self.create(relation_name, data)

        # Return model with new PK value
        return self

    async def delete(self, relation_name: str = None) -> None:
        """Delete this model from the database"""

        # Get the entity of this model instance (which is the metaclass, aka self.__class__)
        entity = self.__class__

        # COMMENT on deleting children
        # Well it could only work for HasOne/MorphOne, or possibly HasMany.  But those tables can simply
        # be deleted manually image.find(1).delete() for example.  For others like BelongsToMany we don't ever want
        # to delete the record and it could be used by many other records through the pivot, so we use unlink() instead
        # and we can also manually delete it if we watned (tags.where(post=1).delete() etc...)
        # So not sure I want a delete() to handle children?  If so maybe just HasOne and MorphOne?  Because with HasMany
        # you would also have to specify WHICH ones to delete, like post.delete('comments', [1,2,3]) etc...
        # With a HasOne/MorphOne you could at least do post.delete('image') and thats quicker than image.find(post=1).delete() manually?

        # Notice I have not yet added delete for HasMany (like post comments).  May be a bit dangerous maybe?

        if relation_name is not None:
            # Deleting children, do NOT delete the parent after the children, this is children only
            field = entity.modelfield(relation_name)
            relation = field.relation.fill(field)
            rel_table = relation.entity.table

            # Notice this will all fail if the child has other constraints
            if type(relation) == HasOne:
                query = (rel_table
                    .delete()
                    .where(getattr(rel_table.c, relation.foreign_key) == getattr(self, entity.pk))
                )
                await self._before_delete()
                await entity.execute(query)
                await self._after_delete()

            elif type(relation) == MorphOne or type(relation) == MorphMany:
                query = (rel_table
                    .delete()
                    .where(getattr(rel_table.c, relation.foreign_type) == entity.tablename)
                    .where(getattr(rel_table.c, relation.foreign_key) == getattr(self, entity.pk))
                )
                await self._before_delete()
                await entity.execute(query)
                await self._after_delete()
            else:
                raise Exception('Deleteing children does not work for this type of relation.')

            # Not sure I should implement OneToMany as you would need to be able to pass in WHICH items to delete
            # but if you go through all that trouble to get the right children models to pass in, you could just
            # delete from the actual child yourself (comments.where(post=1).delete() for example)

        else:
            # Delete this main parent (self) model record from the db
            table = entity.table
            query = table.delete().where(getattr(table.c, entity.pk) == getattr(self, entity.pk))
            await self._before_delete()
            await entity.execute(query)
            await self._after_delete()

    async def link(self, relation_name: str, models: Union[Any, List[Any]]) -> None:
        """Link records to relation using the Many-To-Many pivot table"""

        # NOTICE hooks:  We do NOT actually need to fire hooks on relation tables!
        # Because there are no relation table models to listen for those hooks!

        # NOTICE models:  Models can be single model, single dict, List[Model], List[Dict]
        # and I do NOT have to convert them to actual models.  Because I am using my custom getvalue
        # the pk is pulled regardless of model or dict!

        # Ignore empty models (None or [])
        if not models: return

        # Get the entity of this model instance (which is the metaclass, aka self.__class__)
        entity = self.__class__

        # Get field and relation info
        field = entity.modelfield(relation_name)
        relation = field.relation.fill(field)

        # Ensure models are always a list
        if type(models) != list: models = [models]

        # Insert linkage data one link at a time so we can gracefully skip duplicates
        if type(relation) == BelongsToMany:
            for model in models:
                # Set pivot relation data
                left_key_value = getvalue(self, entity.pk)
                right_key_value = getvalue(model, relation.entity.pk)
                pivot = {
                    relation.left_key: left_key_value,
                    relation.right_key: right_key_value
                }

                # Check if exists.
                # The encode/databases layer does NOT abstract each backend DB libraries exceptions
                # into a common interface so You cannot catch generic IntegrityError.  So instead of a
                # try catch, I will see if the record exists manually first :( - See https://github.com/encode/databases/issues/162
                table = relation.join_table
                query = (
                    sa.select([getattr(table.c, relation.left_key)]).select_from(table)
                    .where(getattr(table.c, relation.left_key) == left_key_value)
                    .where(getattr(table.c, relation.right_key) == right_key_value)
                )
                exists = await entity.fetchone(query)  # Returns None if not exists

                if not exists:
                    query = relation.join_table.insert().values(**pivot)
                    await entity.execute(query)  # No hooks needed, relation table are NOT models, no listeners

                # # Try insert, fail silently if exists
                # try:
                #     query = relation.join_table.insert().values(**pivot)
                #     await entity.execute(query)
                # except:
                #     # Ignore Integrity Errors
                #     pass

        elif type(relation) == MorphToMany:
            for model in models:
                # Set polymorphic pivor relation data
                left_type_value = entity.tablename
                left_key_value = getvalue(self, entity.pk)
                right_key_value = getvalue(model, relation.entity.pk)
                pivot = {
                    relation.left_type: left_type_value,
                    relation.left_key: left_key_value,
                    relation.right_key: right_key_value
                }

                # Check if exists.
                # The encode/databases layer does NOT abstract each backend DB libraries exceptions
                # into a common interface so You cannot catch generic IntegrityError.  So instead of a
                # try catch, I will see if the record exists manually first :( - See https://github.com/encode/databases/issues/162
                table = relation.join_table
                query = (
                    sa.select([getattr(table.c, relation.left_type)]).select_from(table)
                    .where(getattr(table.c, relation.left_type) == left_type_value)
                    .where(getattr(table.c, relation.left_key) == left_key_value)
                    .where(getattr(table.c, relation.right_key) == right_key_value)
                )
                exists = await entity.fetchone(query)  # Returns None if not exists

                if not exists:
                    query = relation.join_table.insert().values(**pivot)
                    await entity.execute(query)  # No hooks needed, relation table are NOT models, no listeners


                # Try insert, fail silently if exists
                # try:
                #     query = relation.join_table.insert().values(**pivot)
                #     await entity.execute(query)
                # except pymysql.err.IntegrityError:
                #     # Ignore Integrity Errors
                #     pass

        else:
            # Linking only works for Many-To-Many relations
            raise Exception('Linking is for Many-To-Many relations only.')

    async def unlink(self, relation_name: str, models: Union[Any, List[Any]] = None) -> None:
        """Unlink records to relation using the Many-To-Many pivot table"""

        # NOTICE hooks:  We do NOT actuall need to fire hooks on relation tables!
        # Because there are no relation table models to listen for those hooks!

        # Get the entity of this model instance (which is the metaclass, aka self.__class__)
        entity = self.__class__

        # Get field and relation info
        field = entity.modelfield(relation_name)
        relation = field.relation.fill(field)

        # Ensure model is a list
        ids = []
        if models is not None:
            if type(models) != list: models = [models]
            ids = [getvalue(x, relation.entity.pk) for x in models]

        if type(relation) == BelongsToMany:
            # Get table and start where on self ID
            table = relation.join_table
            query = table.delete().where(getattr(table.c, relation.left_key) == getattr(self, entity.pk))
            if models is not None:
                # Add in proper relation Ids
                query = query.where(getattr(table.c, relation.right_key).in_(ids))
            await entity.execute(query)  # No hooks needed, relation table are NOT models, no listeners

        elif type(relation) == MorphToMany:
            # Get table and start where on self ID
            table = relation.join_table
            query = (
                table.delete()
                .where(getattr(table.c, relation.left_type) == entity.tablename)
                .where(getattr(table.c, relation.left_key) == getvalue(self, entity.pk))
            )
            if models is not None:
                # Add in proper relation Ids
                query = query.where(getattr(table.c, relation.right_key).in_(ids))
            await entity.execute(query)  # No hooks needed, relation table are NOT models, no listeners
        else:
            raise Exception('Uninking is for Many-To-Many relations only.')

    async def _before_insert(self) -> None:
        """Hook fired before record is inserted (new records only)"""
        event_name = 'uvicore.orm-{' +  self.__class__.modelfqn + '}-BeforeInsert'
        await uvicore.events.codispatch(event_name, {'model': self})

    async def _after_insert(self) -> None:
        """Hook fired after record is inserted (new records only)"""
        event_name = 'uvicore.orm-{' + self.__class__.modelfqn + '}-AfterInsert'
        await uvicore.events.dispatch_async(event_name, {'model': self})

    async def _before_save(self) -> None:
        """Hook fired before record is saved (inserted or updated)"""
        event_name = 'uvicore.orm-{' +  self.__class__.modelfqn + '}-BeforeSave'
        await uvicore.events.codispatch(event_name, {'model': self})

    async def _after_save(self) -> None:
        """Hook fired after record is saved (inserted or updated)"""
        event_name = 'uvicore.orm-{' + self.__class__.modelfqn + '}-AfterSave'
        await uvicore.events.dispatch_async(event_name, {'model': self})

    async def _before_delete(self) -> None:
        """Hook fired before record is deleted"""
        event_name = 'uvicore.orm-{' +  self.__class__.modelfqn + '}-BeforeDelete'
        await uvicore.events.codispatch(event_name, {'model': self})

    async def _after_delete(self) -> None:
        """Hook fired after record is deleted"""
        event_name = 'uvicore.orm-{' + self.__class__.modelfqn + '}-AfterDelete'
        await uvicore.events.dispatch_async(event_name, {'model': self})




# IoC Class Instance
#_ModelIoc: _Model = uvicore.ioc.make('Model', _Model)

# Actual Usable Model Class Derived from IoC Inheritence
#class Model(Generic[E], _BaseModel[E], ModelInterface[E], metaclass=ModelMetaclass):
#class Model(Generic[E], _ModelIoc[E], contracts.Model[E]):
    #pass

# Public API for import * and doc gens
#__all__ = ['_Model', 'Model']



from __future__ import annotations
import uvicore
import sqlalchemy as sa
from dataclasses import dataclass
from uvicore.support import module
from uvicore.support.dumper import dd, dump
from uvicore.contracts import Field as FieldInterface
from uvicore.contracts import Relation as RelationInterface
from typing import Any, Callable, Dict, Optional, OrderedDict

# NOTE dataclass required on each on THAT HAS PROPERTIES event hough they all
# impliment Relation.  If a class does not have property overrides, then do NOT add @dataclass


@dataclass
@uvicore.service()
class Relation(RelationInterface):
    # __slots__ = (
    #     'model',
    #     'foreign_key',
    #     'local_key',
    #     'name',
    #     'entity',
    # )

    # By setting a default on these attributes of a dataclass
    # When printed, only the ones actually set will show up, which makes output much smaller
    model: str
    foreign_key: Optional[str] = None
    local_key: Optional[str] = None
    name: Optional[str] = None
    entity: Optional[Any] = None

    def __init__(self,
        model: str,
        *,
        foreign_key: str = None,
        local_key: str = None,
    ) -> None:
        self.model = model
        self.foreign_key = foreign_key
        self.local_key = local_key

        # Set by self.fill later
        self.name = None
        self.entity = None

    def fill(self, field: Field) -> Relation:
        if not self.foreign_key:
            self.foreign_key = 'id'
        if not self.local_key:
            self.local_key = str(field.name) + '_id'
        self.name = field.name
        self._load_entity()
        return self

    def _fill_reverse(self, field: Field) -> Relation:
        if not self.foreign_key:
            self.foreign_key = str(field.name) + '_id'
        if not self.local_key:
            self.local_key = 'id'
        self.name = field.name
        self._load_entity()
        return self

    def is_one(self) -> bool:
        return type(self) == HasOne or type(self) == BelongsTo or type(self) == MorphOne

    def is_many(self) -> bool:
        return type(self) == HasMany or type(self) == BelongsToMany or type(self) == MorphMany or type(self) == MorphToMany

    def is_type(self, *args) -> bool:
        for arg in args:
            if type(self) == arg:
                return True
        return False

    def contains_many(self, relations: OrderedDict, skip: List = []) -> bool:
        """Walk down all sub_relations by __ name and check if any are *Many"""
        walk = ''
        sub_relations = self.name.split('__')
        i = 0
        for sub_relation in sub_relations:
            # Walk down by concatenation
            walk += sub_relation

            # Skip these sub_relations if desired
            if len(skip) > i and skip[i] == sub_relation:
                walk += '__'
                i += 1
                continue

            # Check relation type for *Many
            relation_type = type(relations.get(walk))
            if relation_type == HasMany or relation_type == BelongsToMany or relation_type == MorphMany or relation_type == MorphToMany:
                return True

            # Walk down
            walk += '__'
            i += 1
        return False

    def _load_entity(self):
        # Fill actual entity class
        if uvicore.ioc.binding(self.model):
            self.entity = uvicore.ioc.make(self.model)
        else:
            self.entity = module.load(self.model).object


@uvicore.service()
class HasOne(Relation):
    """One-To-One Relationship"""

    # In a HasOne, the foreign_key is required since I don't know
    # the name of the current model, so this __init__ override simply
    # makes foreign_key required
    def __init__(self,
        model: str,
        *,
        foreign_key: str,
        local_key: str = None,
    ) -> None:
        super().__init__(model, foreign_key=foreign_key, local_key=local_key)

    # HasOne is also reversed in how it deduces the optional local_key, its just id
    def fill(self, field: Field) -> Relation:
        return self._fill_reverse(field)


@uvicore.service()
class HasMany(Relation):
    """One-To-Many Relationship"""

    # For a HasMany the foreign_key is required
    def __init__(self,
        model: str,
        *,
        foreign_key: str,
        local_key: str = None,
    ) -> None:
        super().__init__(model, foreign_key=foreign_key, local_key=local_key)

    def fill(self, field: Field) -> Relation:
        return self._fill_reverse(field)


@uvicore.service()
class BelongsTo(Relation):
    """Inverse of One-To-One or One-To-Many Relationship"""
    pass


@dataclass
@uvicore.service()
class BelongsToMany(Relation):
    """Many-To-Many Relationship (Both Sides)

    :param model: Related model as import string
    :param join_tablename: Table name of join/intermediate/pivot table
    :param left_key: The foreign key column of the model on which you are defining the relationship
    :param right_key: The foreign key column of the model that you are joining to
    """
    model: str
    join_tablename: str = None
    left_key: str = None
    right_key: str = None
    name: Optional[str] = None
    entity: Optional[Any] = None
    join_table: Optional[sa.Table] = None

    def __init__(self,
        model: str,
        *,
        join_tablename: str,
        left_key: str,
        right_key: str,
    ) -> None:
        self.model = model
        self.join_tablename = join_tablename
        self.left_key = left_key
        self.right_key = right_key

        # Set by self.fill later
        self.name = None
        self.entity = None
        self.join_table = None
        # The left_key is the foreign key name of the model on which you are defining the relationship
        # The right_key is the foreign key name of the model that you are joining to

    def fill(self, field: Field):
        """Fill in additional instance variables like entity and join_table

        :param field: The model field that holds this relation
        """
        # Fill in parameters
        self.name = field.name

        # Load entity
        self._load_entity()

        # Get actual SQLAlchemy relation table
        self.join_table = uvicore.db.table(self.join_tablename, self.entity.connection)
        return self


@dataclass
@uvicore.service()
class Morph(Relation):
    model: str
    polyfix: str = None
    foreign_type: Optional[str] = None
    foreign_key: Optional[str] = None
    local_key: Optional[str] = None
    dict_key: Optional[str] = None
    dict_value: Optional[str] = None
    list_value: Optional[str] = None
    name: Optional[str] = None
    entity: Optional[Any] = None

    def __init__(self,
        model: str,
        polyfix: str = None,
        *,
        foreign_type: str = None,
        foreign_key: str = None,
        local_key: str = None,
        dict_key: str = None,
        dict_value: str = None,
        list_value: str = None,
    ) -> None:
        self.model = model
        self.polyfix = polyfix
        self.foreign_type = foreign_type or polyfix + '_type'
        self.foreign_key = foreign_key or polyfix + '_id'
        self.local_key = local_key or 'id'
        self.dict_key = dict_key
        self.dict_value = dict_value
        self.list_value = list_value

        # Set by self.fill later
        self.name = None
        self.entity = None

    def fill(self, field: Field):
        self.name = field.name
        self._load_entity()
        return self


@uvicore.service()
class MorphOne(Morph):
    pass


@uvicore.service()
class MorphMany(Morph):
    pass


@dataclass
@uvicore.service()
class MorphToMany(Morph):
    model: str
    join_tablename: str = None
    polyfix: str = None
    right_key: str = None
    left_type: Optional[str] = None
    left_key: Optional[str] = None
    dict_key: Optional[str] = None
    dict_value: Optional[str] = None
    list_value: Optional[str] = None
    name: Optional[str] = None
    entity: Optional[Any] = None
    join_table: Optional[sa.Table] = None

    def __init__(self,
        model: str,
        *,
        join_tablename: str,
        polyfix: str,
        right_key: str,
        left_type: str = None,
        left_key: str = None,
        dict_key: str = None,
        dict_value: str = None,
        list_value: str = None,
    ) -> None:
        self.model = model
        self.join_tablename = join_tablename
        self.polyfix = polyfix
        self.right_key = right_key
        self.left_type = left_type or polyfix + '_type'
        self.left_key = left_key or polyfix + '_id'

        # Dict key/value for MorphToMany has NOT yet been implimented
        self.dict_key = dict_key
        self.dict_value = dict_value
        self.list_value = list_value

        # Set by self.fill later
        self.name = None
        self.entity = None
        self.join_table = None

    def fill(self, field: Field):
        # Fill in parameters
        self.name = field.name

        # Load entity
        self._load_entity()

        # Get actual SQLAlchemy relation table
        self.join_table = uvicore.db.table(self.join_tablename, self.entity.connection)
        return self



# Relationship Aliases
# At one point I thought I would make aliases to help other write as OneToOne...
# But this will complicate things and doesn't actually make things any easier
# to read or understand.  Especially because BelongsTo works for ManyToOne and
# also for the inverse side of a OneToOne which I don't have a name for, maybe
# OneToOneInverse = BelongsTo?

# OneToOne        = HasOne
# OneToOneInverse = BelongsTo
# OneToMany       = HasMany
# ManyToOne       = BelongsTo
# ManyToMany      = BelongsToMany
# OneToOneMorph   = MorphOne
# OneToManyMorph  = MorphMany
# ManyToManyMorph = MorphToMany




@dataclass
@uvicore.service()
class Field(FieldInterface):
    column: str
    name: Optional[str] = None
    primary: Optional[bool] = False
    title: Optional[str] = None
    description: Optional[str] = None
    default: Optional[Any] = None
    sortable: Optional[bool] = None
    searchable: Optional[bool] = None
    read_only: Optional[bool] = None
    write_only: Optional[bool] = None
    callback: Optional[Any] = None
    evaluate: Optional[Callable] = None
    relation: Optional[Relation] = None
    json: Optional[bool] = False
    properties: Optional[Dict] = None

    min_length: Optional[int] = None
    max_length: Optional[int] = None
    example: Optional[Any] = None


    # Of all the uvicore Field() arguments, these are VALID OpenAPI fields and handled properly by Pydantic FieldInfo()
    # See pydantic/fields.py def field() or class FieldInfo() for their params.
    # Note that 'required' is not part of FieldInfo but part of pydantics ModelInfo which is automatically
    # handled by the Optional typehinting on the field type.
    # See https://swagger.io/docs/specification/data-models/keywords/ for all valid schema keywords
    # Run your final .json schema through https://apitools.dev/swagger-parser/online/ to validate
    __valid_oepnapi_keywords__ = [
        'title',
        'description',
        'default',
        #'required',
        'example',

        # These get converted to camelCase in metaclass.py for OpenAPI compatibility
        'read_only',
        'write_only',
        'min_length',
        'max_length',
    ]

    # These uvicore Field() arguments will be converted to the x-tra Dict "specification extensions".
    # See https://swagger.io/specification/#specification-extensions
    __convert_to_extensions__ = [
        #'column',
        'sortable',
        'searchable',
        'properties',
    ]

    def __init__(self, column: str = None, *,
        name: Optional[str] = None,
        primary: Optional[bool] = False,
        title: Optional[str] = None,
        description: Optional[str] = None,
        default: Optional[Any] = None,
        #required: Optional[bool] = False,
        sortable: Optional[bool] = None,  # Must be none if not set to hide in OpenAPI
        searchable: Optional[bool] = None,  # Must be none if not set to hide in OpenAPI
        read_only: Optional[bool] = None,  # Must be none if not set to hide in OpenAPI
        write_only: Optional[bool] = None,  # Must be none if not set to hide in OpenAPI
        callback: Optional[Any] = None,
        evaluate: Optional[Callable] = None,
        relation: Optional[Relation] = None,
        json: Optional[bool] = False,
        properties: Optional[Dict] = None,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        example: Optional[Any] = None,
    ):
        self.column = column
        self.name = name
        self.primary = primary
        self.title = title
        self.description = description
        self.default = default
        #self.required = required
        self.sortable = sortable
        self.searchable = searchable
        self.read_only = read_only
        self.write_only = write_only
        self.callback = callback
        self.evaluate = evaluate
        self.relation = relation
        self.json = json
        self.properties = properties
        self.min_length = min_length
        self.max_length = max_length
        self.example = example



from uvicore.typing import Any, AsyncGenerator, Dict, List, Mapping, Optional, Union

import sqlalchemy as sa
from databases import Database as EncodeDatabase
from sqlalchemy.sql import ClauseElement

import uvicore
from uvicore.contracts import Connection
from uvicore.contracts import Package as Package
from uvicore.contracts import Database as DatabaseInterface
from uvicore.database.query import DbQueryBuilder
from uvicore.support.dumper import dd, dump
from sqlalchemy.engine.result import Row as RowProxy


@uvicore.service('uvicore.database.db.Db',
    aliases=['Database', 'database', 'db'],
    singleton=True,
)
class Db(DatabaseInterface):
    """Database private class.

    Do not import from this location.
    Use the uvicore.db singleton global instead."""

    @property
    def default(self) -> str:
        return self._default

    @property
    def connections(self) -> Dict[str, Connection]:
        return self._connections

    @property
    def engines(self) -> Dict[str, sa.engine.Engine]:
        return self._engines

    @property
    def databases(self) -> Dict[str, EncodeDatabase]:
        return self._databases

    @property
    def metadatas(self) -> Dict[str, sa.MetaData]:
        return self._metadatas

    def __init__(self) -> None:
        self._default = None
        self._connections = Dict()
        self._engines = Dict()
        self._databases = Dict()
        self._metadatas = Dict()

    def init(self, default: str, connections: Dict[str, Connection]) -> None:
        self._default = default
        self._connections = connections

        # For each unique metakey, create engines, encode databases and metadatas
        for connection in connections.values():
            # Check if we have already handled this unique metakey
            if connection.metakey in self.metadatas: continue

            if connection.backend == 'sqlalchemy':
                # Build encode/databases specific connection URL
                # connection.url has a dialect in it, which we need for engines
                # but don't need for encode/databases library
                if connection.driver == 'sqlite':
                    encode_url = (connection.driver
                        + ':///' + connection.database
                    )
                else:
                    encode_url = (connection.driver
                        + '://' + connection.username
                        + ':' + connection.password
                        + '@' + connection.host
                        + ':' + str(connection.port)
                        + '/' + connection.database
                    )
                self._engines[connection.metakey] = sa.create_engine(connection.url)
                self._databases[connection.metakey] = EncodeDatabase(encode_url)
                self._metadatas[connection.metakey] = sa.MetaData()

    def packages(self, connection: str = None, metakey: str = None) -> List[Package]:
        if not metakey:
            if not connection: connection = self.default
            metakey = self.connection(connection).metakey
        packages = []
        for package in uvicore.app.packages.values():
            if not 'database' in package: continue
            for conn in package.database.connections.values():
                if conn.metakey == metakey:
                    packages.append(package)
        return packages

    def metakey(self, connection: str = None, metakey: str = None) -> str:
        try:
            if not metakey:
                if not connection:
                    connection = self.default
                metakey = self.connection(connection).metakey
            return metakey
        except Exception:
            dump("ERROR Connections:", self.connections)
            raise Exception('Metakey not found, connection={} metakey={}'.format(connection, metakey))

    def connection(self, connection: str = None) -> Connection:
        if not connection: connection = self.default
        return self.connections.get(connection)

    def metadata(self, connection: str = None, metakey: str = None) -> sa.MetaData:
        metakey = self.metakey(connection, metakey)
        return self.metadatas.get(metakey)

    def tables(self, connection: str = None, metakey: str = None) -> List[sa.Table]:
        metadata = self.metadata(connection, metakey)
        return metadata.tables

    def table(self, table: str, connection: str = None) -> sa.Table:
        tablename = self.tablename(table, connection)
        metadata = self.metadata(connection)
        if metadata: return metadata.tables.get(tablename)

    def tablename(self, table: str, connection: str = None) -> str:
        if '.' in table:
            connection, table = tuple(table.split('.'))
        connection = self.connection(connection)
        if connection:
            return connection.prefix + table

    def engine(self, connection: str = None, metakey: str = None) -> sa.engine.Engine:
        metakey = self.metakey(connection, metakey)
        return self.engines.get(metakey)

    async def database(self, connection: str = None, metakey: str = None) -> EncodeDatabase:
        metakey = self.metakey(connection, metakey)

        # To connect on-the-fly or NOT on-the-fly, in the service uvicore_startup event, or both???

        # On-The-Fly ON Only
        # - When running from ./serve-uvicorn and using wrk -c10 -t4 -d5 http://0.0.0.0:5000/api/tags/1
        #   Notice the 2 IDs 0x7f733acc9130 and 0x7f733acc96d0 are different
        #     [ERROR] Exception in ASGI application
        #     File "/home/mreschke/.cache/pypoetry/virtualenvs/mreschke-speedtest-epfwGmSK-py3.9/lib/python3.9/site-packages/databases/backends/mysql.py", line 100, in release
        #       await self._database._pool.release(self._connection)
        #     File "/home/mreschke/.cache/pypoetry/virtualenvs/mreschke-speedtest-epfwGmSK-py3.9/lib/python3.9/site-packages/aiomysql/pool.py", line 204, in release
        #       assert conn in self._used, (conn, self._used)
        #     AssertionError: (<aiomysql.connection.Connection object at 0x7f733acc9130>, {<aiomysql.connection.Connection object at 0x7f733acc96d0>})

        # On-The-Fly OFF (using Service events Only)
        # - If off, during TESTING, uvicore.console.events.command.Startup event does not fire, so DB
        #   is not connected and all tests against DB fail (can't seed tables, canot query tables..)

        # Solution
        # Alter test suite to fire a uvicore.console.events.command.PytestStartup and add that event to db/services.py uvicore_startup handler

        # Connect on-the-fly - NO, NOT required anymore
        # self.connect(connection, metakey)

        # Return Encode database
        return self.databases.get(metakey)

    async def connect(self, connection: str = None, metakey: str = None, *, all_dbs: bool = False) -> None:
        if all_dbs:
            # Connect all DBs
            for metakey, database in uvicore.db.databases.items():
                if not database.is_connected:
                    try:
                        await database.connect()
                    except:
                        # Will catch if DB is not running or port/constr is wrong
                        pass
        else:
            # Connect a single DB
            metakey = self.metakey(connection, metakey)
            database = self.databases.get(metakey)  # Dont call self.database() as its recursive
            if not database.is_connected:
                try:
                    await database.connect()
                except:
                    # Will catch if DB is not running or port/constr is wrong
                    pass

    async def disconnect(self, connection: str = None, metakey: str = None, all_dbs: bool = False) -> None:
        if all_dbs:
            # Disconnect from all connected databases
            for database in self.databases.values():
                # Only disconnect if connected or will throw an error
                if database.is_connected:
                    await database.disconnect()
        else:
            # Disconnect from one database by connection str or metakey
            metakey = self.metakey(connection, metakey)
            database = self.databases.get(metakey)
            if database.is_connected:
                await database.disconnect()

    async def fetchall(self, query: Union[ClauseElement, str], values: Dict = None, connection: str = None, metakey: str = None) -> List[RowProxy]:
        database = await self.database(connection, metakey)
        return await database.fetch_all(query, values)

    async def fetchone(self, query: Union[ClauseElement, str], values: Dict = None, connection: str = None, metakey: str = None) -> Optional[RowProxy]:
        database = await self.database(connection, metakey)
        return await database.fetch_one(query, values)

    async def execute(self, query: Union[ClauseElement, str], values: Union[List, Dict] = None, connection: str = None, metakey: str = None) -> Any:
        database = await self.database(connection, metakey)
        if type(values) == dict:
            return await database.execute(query, values)
        elif type(values) == list:
            return await database.execute_many(query, values)
        else:
            return await database.execute(query)

    #  async def iterate(self, query: Union[ClauseElement, str], values: dict = None, connection: str = None) -> AsyncGenerator[Mapping, None]:
    #     async with self.connection() as connection:
    #         async for record in connection.iterate(query, values):
    #             yield record

    def query(self, connection: str = None) -> DbQueryBuilder[DbQueryBuilder, Any]:
        if not connection: connection = self.default
        return DbQueryBuilder(connection)


# IoC Class Instance
# Not to be imported by the public from here.
# Use the uvicore.db singleton global instead.



import uvicore
from uvicore.package import Provider
from uvicore.http.package import bootstrap
from uvicore.support.dumper import dump, dd
from uvicore.http.package.registers import Http
from uvicore.console.package.registers import Cli
from uvicore.foundation.events import app as AppEvents


@uvicore.provider()
class Http(Provider, Cli, Http):

    def register(self) -> None:
        """Register package into the uvicore framework.
        All packages are registered before the framework boots.  This is where
        you define your packages configs, IoC bindings and early event listeners.
        Configs are deep merged only after all packages are registered.  No real
        work should be performed here as it is very early in the bootstraping
        process and we have no clear view of the full configuration system."""

        # Register events used in this package
        # NO, moved to events/server.py using @uvicore.event() to auto-register event list.
        # self.events.register(
        #     name='uvicore.http.events.server.Startup',
        #     description='HTTP Server has been started.  This is the Starlette startup event.',
        #     is_async=False,
        # )
        # self.events.register(
        #     name='uvicore.http.events.server.Shutdown',
        #     description='HTTP Server has been shutdown.  This is the Starlette shutdown event.',
        #     is_async=False,
        # )

        #     @uvicore.app.http.on_event("shutdown")
        #     async def shutdown():
        #         for database in self.databases.values():
        #             await database.disconnect()

        # Register event listeners
        # After all providers are booted we have a complete list of view paths
        # and template options fully merged.  Now we can fire up the static
        # paths and template system.
        # We handle this event even in CONSOLE mode to handle edge cases.
        AppEvents.Booted.listen(bootstrap.Http)


    def boot(self) -> None:
        """Bootstrap package into the uvicore framework.
        Boot takes place after ALL packages are registered.  This means all package
        configs are deep merged to provide a complete and accurate view of all
        configuration. This is where you register, connections, models,
        views, assets, routes, commands...  If you need to perform work after ALL
        packages have booted, use the event system and listen to the booted event:
        self.events.listen('uvicore.foundation.events.app.Booted', self.booted)"""

        # Define service provider registration control
        self.registers(self.package.config.registers)

        # Define HTTP Middleware
        #self.middleware(uvicore.config('app.middleware'))

        # Define CLI commands
        self.register_cli_commands({
            'http': {
                'help': 'Uvicore HTTP Commands',
                'commands': {
                    'serve': 'uvicore.http.commands.serve.cli',
                    'routes': 'uvicore.http.commands.routes.list',
                },
            },

            # Extend schematic generator commands
            'gen': {
                'commands': {
                    'controller': 'uvicore.http.commands.generators.controller',
                    'api-controller': 'uvicore.http.commands.generators.api_controller',
                    'composer': 'uvicore.http.commands.generators.composer',
                },
            }
        })

        # Uvicore default template options.  Because this HTTP package is high up
        # your package can easily override any of these default options!
        from uvicore.http.templating import context_functions
        self.register_http_view_context_processors({
            'context_functions': {
                'url': context_functions.url,
                'asset': context_functions.asset,
                'public': context_functions.public,
            },
        })




# type: ignore
from .db import Db
from .table import Table
from uvicore.contracts import Connection as ConnectionInterface


#from .connection import Connection
# NO, just import it to keep the same path
# Proxy instead of import to creata a nicer uvicore.database.Connection namespace
# Even if you override this the acutal class, all listings will still show this consistent namespace
#from uvicore.contracts import Connection as ConnectionInterface
# class Connection(ConnectionInterface):
#     pass

# Connection was in its own file, but I hate seeing uvicore.database.connection.Connection
# in all my dumps.  So I just added the class here directly
class Connection(ConnectionInterface):
    pass



import uvicore
from uvicore.http import status
from uvicore.http import Request
from uvicore.auth import User
from uvicore.support import module
from fastapi.params import Security
from fastapi.security import SecurityScopes
from uvicore.support.dumper import dump, dd
from uvicore.typing import Optional, Sequence, Callable, Any, List, Dict, Tuple
from uvicore.http.exceptions import HTTPException, PermissionDenied, NotAuthenticated, InvalidCredentials
from uvicore.contracts import UserProvider


@uvicore.service()
class Guard(Security):
    """Uvicore Auth Guard"""

    def __init__(self, scopes: Optional[Sequence[str]] = None, guard: str = None):
        # # Swap guard and scopes
        # if scopes is None and type(guard) == list:
        #     scopes: Sequence[str] = guard
        #     guard = None

        # Ensure scopes is a List, to allow for singles
        if scopes: scopes = [scopes] if isinstance(scopes, str) else list(scopes)

        # Do NOT apply a default guard to self.guard, let it be None
        # So I know its blank so I can overwrite it with a parent guard if needed
        self.scopes = scopes
        self.guard = guard

        # Get auth_config from app config
        auth_config = uvicore.config.app.auth

        # Set default guard if none providedGet actual guard from app config
        if guard is None: guard = auth_config.default

        # Get actual guard config
        if guard not in auth_config.guards:
            raise Exception('Guard {} not found in app config'.format(guard))
        guard_config = auth_config.guards[guard].clone()  # Clone becuase I add name below

        # Add name to guard config (its a clone, its ok)
        guard_config.name = guard

        # Get all providers from auth_config
        providers = auth_config.providers

        # Get all authenticator options from auth_config
        options = auth_config.options

        # Call parent Depends passing in the multi-middleware Authenticator
        super().__init__(dependency=Authenticator(guard_config, options, providers), scopes=scopes, use_cache=True)


@uvicore.service()
class Authenticator:
    def __init__(self, guard: Dict, options, providers: Dict):
        # Guard is the full config SuperDict from app config matching the proper guard string
        self.guard = guard

        # Authenticator Default Options
        self.options = options

        # Providers is all providers from app config
        self.providers = providers

    async def __call__(self, scopes: SecurityScopes, request: Request):
        #dump(self.guard)
        # Dict({
        #     'authenticators': Dict({
        #         'jwt': Dict({
        #             'module': 'uvicore.auth.middleware.Jwt',
        #             'verify_signature': True,
        #             'algorithms': ['RS256'],
        #             'secret':
        #                 '-----BEGIN PUBLIC KEY-----\nMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAnc84SDViVX8JNye2GVQZ\n'
        #                 'ixAwG2PWXoOhkj++wGASoAXs2LN0Ue48conxf/0bgEtq6kcbLPR23SieqBZA77vc\n'
        #                 'yulimMbzfwNczyP3FRo8wSCqRgJipTse87WItd8ga2MUCzSS8q19V4swUT4T23Su\n'
        #                 'DiG/Ry5f1sYbvxP2kJAJMUCzVbS7STxh33h65Bj+P6JdzrCJi+yrLqg928RHjLIF\n'
        #                 'gDy4MyFBLTI8w5u6IJi1TLm6h9lj3YqSa/qDkkIardnnZa7Xj0IJCEB9c+RD4Q7C\n'
        #                 '+jco6g2Vr9oLP8Mg3c5lZPNVzcXC67UMVk9lK+zrlfPDI/m2+9kyTc/58S9ZUTFJ\nQwIDAQAB\n-----END PUBLIC KEY-----'
        #         }),
        #         'basic': Dict({
        #             'module': 'uvicore.auth.middleware.Basic',
        #             'provider': 'users',
        #             'realm': 'App1'
        #         })
        #     }),
        #     'name': 'api'
        # })

        for authenticator in self.guard.authenticators.values():
            # Get authenticator options by deep merging defaults and proper providers
            options = authenticator.clone()
            if 'options' in options:
                # Deep merge default options
                option_key = options.options
                if option_key not in self.options:
                    # This is an application error, not an HTTPException
                    raise Exception('Default options key {} not found in app config'.format(option_key))
                options.defaults(self.options[option_key])  # Merge seems to do a clone too!

            # Merge provider into options
            if 'provider' in options:
                if options.provider not in self.providers:
                    # This is an application error, not an HTTPException
                    raise Exception('Provider {} not found in app config'.format(options.provider))
                options.provider = self.providers[options.provider].clone()

            options.guard = self.guard.name

            #dump(options)
            # Dict({
            #     'default_options': 'jwt',
            #     'module': 'uvicore.auth.middleware.Jwt',
            #     'provider': Dict({
            #         'module': 'uvicore.auth.models.user.User',
            #         'method': 'userinfo',
            #         'model': 'uvicore.auth.models.user.User',
            #         'includes': ['roles', 'roles.permissions', 'groups', 'groups.roles', 'groups.roles.permissions']
            #     }),
            #     'sync': Dict({'auto_create_user': True}),
            #     'verify_signature': True,
            #     'audience': '222b06eb-85ce-472b-af30-ec09244e3bf0',
            #     'algorithms': ['RS256'],
            #     'secret':
            #         '-----BEGIN PUBLIC KEY-----\nMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAnc84SDViVX8JNye2GVQZ\n'
            #         'ixAwG2PWXoOhkj++wGASoAXs2LN0Ue48conxf/0bgEtq6kcbLPR23SieqBZA77vc\n'
            #         'yulimMbzfwNczyP3FRo8wSCqRgJipTse87WItd8ga2MUCzSS8q19V4swUT4T23Su\n'
            #         'DiG/Ry5f1sYbvxP2kJAJMUCzVbS7STxh33h65Bj+P6JdzrCJi+yrLqg928RHjLIF\n'
            #         'gDy4MyFBLTI8w5u6IJi1TLm6h9lj3YqSa/qDkkIardnnZa7Xj0IJCEB9c+RD4Q7C\n'
            #         '+jco6g2Vr9oLP8Mg3c5lZPNVzcXC67UMVk9lK+zrlfPDI/m2+9kyTc/58S9ZUTFJ\nQwIDAQAB\n-----END PUBLIC KEY-----'
            # })

            # Import the auth middleware module
            middleware = module.load(options.module).object(options)

            # Fire the middleware __call__ callable and get the returned value
            value = await middleware(scopes, request)

            # If value is returned, auth was successful with this authenticator.  Return value, stop the middleware stack.
            # If value is None means auth headers not found.  Continue to next middleware in auth stack
            if value is not None:
                return value

        # If we are here, no auth middleware returned a value, meaning NOT logged in
        # If no value is ever returned we are not logged in.
        # Maybe here is the place to add an anonymous user to the request? with user.authenticated = False ???????
        # NO, think about this.  If a route is NOT guarded, this code will never run therefore request.user will
        # never exist.  Would have to do global auth middleware to accomplish an always present anonymous user.
        # I could have a built-in hidden global middleware that adds request.user as an anonymous model?
        raise NotAuthenticated('MASTER STACKER')


@uvicore.service()
class Auth:
    """Base Auth middleware class"""

    async def retrieve_user(self, username: str, password: str, provider: Dict) -> Optional[User]:
        """Retrieve user from User Provider backend"""
        # Import our user provider defined in auth config
        user_provider: UserProvider = module.load(provider.module).object()

        # Get user from user provider and validate password
        # If returned user is None, validation has failed, user is disabled or user not found
        user = await user_provider.retrieve_by_credentials(username, password, **provider.options)

        # Do not throw error if no user or not validated here.  We let the middleware handle that
        return user

    def validate_permissions(self, user: User, scopes: SecurityScopes) -> None:
        """Validate logged in users permissions again route permissions"""

        # Superadmin is always allowed
        if user.superadmin: return

        # Get permissions defined on this route
        route_permissions = scopes.scopes

        # If route does not specify permissions, then anyone that is authenticated can access.
        if not route_permissions: return

        # Compare users permissions with route permissions
        for permission in route_permissions:
            if permission in user.permissions:
                # This is an OR, if any one of these, then pass
                return

        # No matching permissinos means they are logged in, but they don't have the proper permissions.
        raise PermissionDenied(route_permissions)

    def auth_header(self, request) -> Tuple[str, str, str]:
        """Extract authorization header parts"""
        authorization = request.headers.get('Authorization')
        if not authorization: return (authorization, '', '')
        scheme, _, param = authorization.partition(' ')
        return authorization, scheme.lower(), param



import uvicore
from uvicore.console import command, argument, option
from uvicore.support.dumper import dd, dump


@command()
@option('--raw', is_flag=True, help='Show output without prettyprinter')
def bindings(raw: bool = False):
    """List all Ioc Bindings"""
    if not raw:
        uvicore.log.header("List of all Ioc bindings")
        uvicore.log.line()
        dump(uvicore.ioc.bindings)
    else:
        dic = {}
        for key, binding in uvicore.ioc.bindings.items():
            dic[key] = binding.__dict__
        print(dic)


@command()
def singletons():
    """List Singleton Ioc Bindings"""
    uvicore.log.header("List of all Ioc singleton bindings")
    uvicore.log.line()
    bindings = {key:binding for (key, binding) in uvicore.ioc.bindings.items() if binding.singleton == True}
    dump(bindings)


@command()
@argument('type')
def type(type: str):
    """List Ioc Bindings of a Specific Type (comma separated)"""
    uvicore.log.header("List of all {} Ioc bindings".format(type))
    uvicore.log.line()

    types = []
    if ',' in type:
        parts = type.split(',')
        for part in parts:
            types.append(part.upper())
    else:
        types = [type.upper()]

    bindings = {}
    for key, binding in uvicore.ioc.bindings.items():
        if binding.type.upper() in types:
            bindings[key] = binding
    #bindings = {key:binding for (key, binding) in uvicore.ioc.bindings.items() if binding.type.upper() == type}
    dump(bindings)


@command()
def overrides():
    """List Overridden Ioc Bindings"""
    uvicore.log.header("List of all Ioc bindings that have been overridden")
    uvicore.log.line()
    bindings = {key:binding for (key, binding) in uvicore.ioc.bindings.items() if binding.path != key}
    # overridden = {}
    # for key, binding in uvicore.ioc.bindings.items():
    #     if key != binding.path:
    #         overridden[key] = binding
    dump(bindings)


@command()
@argument('key', default='')
@option('--raw', is_flag=True, help='Show output without prettyprinter')
def get(key: str = None, raw: bool = False):
    """Get a binding by name"""
    if not raw:
        dump(uvicore.ioc.binding(key))
    else:
        print(uvicore.ioc.binding(key).__dict__)



# type: ignore
import sys

# If using installed version of asyncclick
#import asyncclick as click
#from asyncclick import argument, option

# If using local copy of asyncclick
from . import asyncclick as click
from .asyncclick import argument, option


from .decorators import command, group

# Does work without error, but never accessed by a user, so no point
# No from .console import cli

# The console package uses a __init__.py because users will import these
# methods from their own commands and we want a nicer import which looks
# like - from uvicore.console import command


def command_is(command: str):
    command = command.lower()
    running = ' '.join(sys.argv[1:]).lower()
    #return len(sys.argv) > 1 and command.lower() == running.lower()
    return command in running



import os

def find_base(file):
    """Finds app base path relative to some __file__ in the app"""
    count = 0
    path=os.path.dirname(os.path.realpath(file))
    while count < 50:
        if (
            os.path.exists(os.path.realpath(path + '/uvicore')) or
            os.path.exists(os.path.realpath(path + '/setup.py')) or
            os.path.exists(os.path.realpath(path + '/.env'))
        ):
            return os.path.realpath(path)
        else:
            path = os.path.realpath(path + '/../')
        count += 1
    exit("Could not find base path")



import uvicore
from uvicore.events import Event


@uvicore.event()
class Registered(Event):
    """Application bootstrap has registered all package service providers."""

    is_async = False

    def __init__(self):
        self.x = 'x here'
        pass


@uvicore.event()
class Booted(Event):
    """Application bootstrap has booted all package service providers."""

    is_async = False

    def __init__(self):
        pass


# Register events used in this package
# self.events.register(
#     name='uvicore.foundation.events.app.Registered',
#     description='Application bootstrap has registered all package service providers.',
#     dynamic=False,
#     is_async=False,
# )
# self.events.register(
#     name='uvicore.foundation.events.app.Booted',
#     description='Application bootstrap has booted all package service providers.',
#     dynamic=False,
#     is_async=False,
# )



import glob
from importlib import import_module
from importlib.util import find_spec
from dataclasses import dataclass
from uvicore.support.dumper import dump, dd
from typing import Any, NamedTuple

@dataclass
class Module():
    """Dynamic imported module interface"""
    object: Any
    name: str
    path: str
    fullpath: str
    package: str
    file: str

def load(module: str) -> Module:
    """Import module from string"""
    # Detect if wildcard some.module.*
    # If so, we will import EACH file in that directory
    # Meaning no __init__.py required to load all
    wildcard = False
    if module[-2:] == '.*':
        wildcard = True
        module = module[0:-2]

    # Explode parts
    parts = module.split('.')
    path = '.'.join(parts[0:-1])
    name = ''.join(parts[-1:])

    # Try to import assuming module string is an object, a file or a package (with __init__.py)
    #try:
    # Namespace means you are importing a folder without a __init__.py
    namespace = False
    root = False

    # Root means you are importing a single module without a path
    # like just 'uvicore'
    if path == '': root = True

    if namespace or root:
        imported = import_module(module)
    else:
        try:
            imported = import_module(module)
            namespace = True
        except ModuleNotFoundError:
            imported = import_module(path)

    # Example when importing an actual dictiony called app
    # from uvicore.foundation.config.app.app
    # imported.__name__ # uvicore.foundation.config.app
    # imported.__package__ # uvicore.foundation.config
    # imported.__file__ # /home/mreschke/Code...

    # Get actual imported object
    if namespace or root:
        object = imported
    else:
        #try:
        #dump('hi', imported)
        object = getattr(imported, name)
        #except:
        #    raise Exception("There was an error while dynamically importing " + module + '.  Check for errors in that file.')

    # File can be actual file.py or __init__.py or just the folder
    # if its a namespace import
    file = imported.__file__ or imported.__path__._path[0]

    if wildcard:
        pyfiles = glob.glob(file + '/*.py')
        for pyfile in pyfiles:
            # Recursively load each .py file in this folder
            modname = pyfile.split('/')[-1].split('.')[0]
            load(module + '.' + modname)
        # No need to load actual namespace module since we load .*
        return

    # Build our Module() object for return
    mod = Module(
        object=object,
        name=name,
        path=path if path else name,
        fullpath=path + '.' + name if path else name,
        package=imported.__package__,
        file=file
    )
    return mod
    # except:
    #     raise ModuleNotFoundError("Could not dynamically load module {}".format(module))

def location(module: str) -> str:
    """Find modules folder path (not file path) without importing it"""

    try:
        # Module is a package or namespace package
        spec = find_spec(module)
        return spec.submodule_search_locations[0]
    except:
        # Module is actual file or class/method inside a file
        try:
            spec = find_spec('.'.join(module.split('.')[0:-1]))
            if spec.submodule_search_locations:
                # Module is actual file
                return spec.submodule_search_locations[0]
            else:
                # Module is a class/method inside a file
                spec = find_spec('.'.join(module.split('.')[0:-2]))
                return spec.submodule_search_locations[0]
        except:
            return None



    # if spec.submodule_search_locations[0]:
    #     return spec.submodule_search_locations[0]
    # else:
    #     return spec.origin



from uvicore.configuration import env
from uvicore.typing import OrderedDict

# This is the main auth config.  All items here can be overridden
# when used inside other applications.  Accessible at config('uvicore.auth')

config = {

    # --------------------------------------------------------------------------
    # Argon2 Password Hasher Configuration
    # See https://argon2-cffi.readthedocs.io/en/stable/api.html
    # --------------------------------------------------------------------------
    'hasher': {
        'type': 'ID',
        'hash_len': 16,
        'salt_len': 16,
        'encoding': 'utf-8',
        'time_cost': 2,
        'memory_cost': 102400,
        'parallelism': 8,
    },


    # --------------------------------------------------------------------------
    # Web Configuration
    #
    # prefix: All web routes will be prefixed with this URI. Ex: '' or '/wiki'
    #         This is in addition to running apps web.prefix config
    # --------------------------------------------------------------------------
    'web': {
        'prefix': '/auth',
    },

    # --------------------------------------------------------------------------
    # Api Configuration
    #
    # prefix: All api routes will be prefixed with this URI. Ex: '' or '/wiki'
    #         This is in addition to running apps api.prefix config
    # --------------------------------------------------------------------------
    'api': {
        'prefix': '/auth',
    },


    # --------------------------------------------------------------------------
    # Database Connections
    # --------------------------------------------------------------------------
    'database': {
        'default': 'auth',
        'connections': {
            'auth': {
                'backend': env('DB_AUTH_BACKEND', 'sqlalchemy'),
                'driver': env('DB_AUTH_DRIVER', 'mysql'),
                'dialect': env('DB_AUTH_DIALECT', 'pymysql'),
                'host': env('DB_AUTH_HOST', '127.0.0.1'),
                'port': env.int('DB_AUTH_PORT', 3306),
                'database': env('DB_AUTH_DB', 'auth'),
                'username': env('DB_AUTH_USER', 'root'),
                'password': env('DB_AUTH_PASSWORD', 'techie'),
                'prefix': env('DB_AUTH_PREFIX', None),

                #'driver': 'sqlite',
                #'database': ':memory',
                #'database': '/tmp/x.db',
                #'prefix': None,
            },
        },
    },


    # --------------------------------------------------------------------------
    # Registration Control
    # --------------------------------------------------------------------------
    # This lets you control the service provider registrations.  If this app
    # is used as a package inside another app you might not want some things
    # registered in that context.
    # 'registers': {
    #     'web_routes': True,
    #     'api_routes': True,
    #     'middleware': True,
    #     'views': True,
    #     'assets': True,
    #     'commands': True,
    #     'models': True,
    #     'tables': True,
    #     'seeders': True,
    # },


    # --------------------------------------------------------------------------
    # Package Dependencies (Service Providers)
    #
    # Define all the packages that this package depends on.  At a minimum, only
    # the uvicore.foundation package is required.  The foundation is very
    # minimal and only depends on configuratino, logging and console itself.
    # You must add other core services built into uvicore only if your package
    # requires them.  Services like uvicore.database, uvicore.orm, uvicore.http
    # uvicore.auth...
    # --------------------------------------------------------------------------
    # 'dependencies': OrderedDict({})

}



import uvicore
import aioredis
from uvicore.typing import Dict, Any
from uvicore.support.dumper import dump, dd

# Basically a uvicore quick connect and passthrough of aioredis
# For aioredis commands see https://aioredis.readthedocs.io/en/latest/mixins.html#generic-commands

@uvicore.service('uvicore.redis.redis.Redis',
    aliases=['Redis', 'redis'],
    singleton=True,
)
class Redis:

    @property
    def default(self) -> str:
        return self._default

    @property
    def connections(self) -> Dict[str, str]:
        return self._connections

    @property
    def engines(self) -> Dict:
        """All connected redis pools"""
        return self._engines

    def __init__(self) -> None:
        self._default = None
        self._connections = Dict()
        self._engines = Dict()

    def init(self, default: str, connections: Dict[str, str]):
        self._default = default
        self._connections = connections
        for connection in self.connections.values():
            connection.url = (
                'redis://'
                + connection.host + ':'
                + str(connection.port) + '/'
                + str(connection.database)
            )
            if connection.password:
                connection.url += '?password=' + connection.password

    def connection(self, connection: str = None):
        """Get one connection by name"""
        if not connection: connection = self.default
        conn = self.connections.get(connection)
        if not conn:
            raise Exception('Redis connection {} not found'.format(connection))
        return conn

    async def connect(self, connection: str = None) -> aioredis.Redis:
        """Connect to a redis database by uvicore connection"""
        conn = self.connection(connection)

        # Connect to redis pool if connection never started
        if conn.url not in self.engines:
            self._engines[conn.url] = await aioredis.create_redis_pool(conn.url)

        # Return actual connection (engine)
        return self.engines[conn.url]




    # def _connection(self, connection: str = None):
    #     """Get one connection by name"""
    #     if not connection: connection = self._current_conn or self.default
    #     self._current_conn = None
    #     conn = self.connections.get(connection)
    #     if not conn:
    #         raise Exception('Redis connection {} not found'.format(connection))
    #     return conn

    # def connection(self, connection: str = None):
    #     """Change the default connection for one query"""
    #     if connection: self._current_conn = connection
    #     return self

    # async def connect(self, connection: str = None):
    #     conn = self._connection(connection)

    #     # Connect to redis if connection never started
    #     if conn.url not in self.engines:
    #         self.engines[conn.url] = await aioredis.create_redis_pool(conn.url)

    #     # Return actual connection (engine)
    #     return self.engines[conn.url]

    # async def get(self, key: Any, default: Any = None):
    #     redis = await self.connect()
    #     return await redis.get(key)

    # async def set(self, key: Any, value: Any):
    #     redis = await self.connect()
    #     return await redis.set(key, value)



class hybridmethod(classmethod):
    """Decorator for a class method that can be both a @classmethod or regular instance method"""
    def __get__(self, instance, type_):
        descr_get = super().__get__ if instance is None else self.__func__.__get__
        return descr_get(instance, type_)



import uvicore
from uvicore.support import module
from uvicore.typing import Dict, Any
from uvicore.support.dumper import dump, dd
from uvicore.contracts import Cache as CacheInterface


@uvicore.service('uvicore.cache.manager.Manager',
    aliases=['Cache', 'cache'],
    singleton=True,
)
class Manager:

    @property
    def default(self) -> str:
        return self._default

    @property
    def stores(self) -> Dict[str, Dict]:
        return self._stores

    @property
    def backends(self) -> Dict:
        return self._backends

    def __init__(self):
        #self.default = config.default
        #self.stores = config.stores
        self._current_store = None
        self._backends = Dict()

        # Cache app config is optional
        config = uvicore.config.app.cache.defaults({
            'default': 'array',
            'stores': {
                'redis': {
                    'driver': 'uvicore.cache.backends.redis.Redis',
                    'connection': 'cache',
                    'prefix': 'uvicore.cache::cache/',
                    'seconds': 600,
                },
                'array': {
                    'driver': 'uvicore.cache.backends.array.Array',
                    'prefix': 'uvicore.cache::cache/',
                    'seconds': 60,
                },
            }
        })
        self._default: str = config.default
        self._stores: Dict = config.stores

    def connect(self, store: str = None) -> CacheInterface:
        """Connect to a cache backend store"""
        store_name = store or self.default
        store = self.stores.get(store_name)
        if not store:
            raise Exception('Cache store {} not found'.format(store_name))

        if store_name not in self.backends:
            # Instantiate, connect and save store in local backends cache
            driver = module.load(store.driver).object(self, store)
            self._backends[store_name] = driver

        return self.backends[store_name]

    def store(self, store: str = None) -> CacheInterface:
        """Alias to connect"""
        return self.connect(store)



import uvicore
from uvicore.auth.models.role import Role
from uvicore.support.dumper import dump, dd
from uvicore.auth.models.permission import Permission


@uvicore.seeder()
async def seed():
    uvicore.log.item('Seeding table roles')

    # Get all permissions
    perms = await Permission.query().key_by('name').get()

    await Role.insert_with_relations([
        {
            'name': 'Administrator',
            'permissions': [
                perms.get('admin'),
            ]
        },
        {
            'name': 'User',
        },
        {
            'name': 'Anonymous',
        },
    ])

    # await Role(name='Administrator').save()

    # user = await Role(name='Users').save()
    # await user.link('permissions', [
    #     perms['attributes.create'],
    #     perms['attributes.read'],
    #     perms['attributes.update'],
    #     perms['attributes.delete'],
    # ])

